import json
from typing import Dict, List, Optional, TypedDict, cast

from more_itertools import pairwise

from kolo.types import JSON


class OutboundRequest(TypedDict):
    body: str
    headers: Dict[str, str]
    method: str
    method_and_full_url: str
    url: str


class OutboundResponse(TypedDict):
    body: Optional[str]
    headers: Optional[Dict[str, str]]
    method: str
    method_and_full_url: str
    status_code: Optional[int]
    url: str


class OutboundJSONResponse(OutboundResponse, total=False):
    content_type: Optional[str]
    json_body: JSON


class OutboundFrame(TypedDict):
    request: OutboundRequest
    response: OutboundJSONResponse


def update_response_json_body(response: OutboundJSONResponse) -> OutboundJSONResponse:
    if response["headers"] is None:
        return response

    content_type = (
        response["headers"].get("Content-Type")
        or response["headers"].get("content-type")
        or response["headers"].get("CONTENT-TYPE")
    )
    response["content_type"] = content_type
    content_type_is_json = content_type and "application/json" in content_type
    body = response.get("body", None)
    if content_type_is_json and isinstance(body, str):
        response["json_body"] = json.loads(body)
    return response


def parse_outbound_frames(frames) -> List[OutboundFrame]:
    frames = (
        frame
        for frame in frames
        if frame["type"] in ("outbound_http_request", "outbound_http_response")
    )

    without_urllib3 = (frame for frame in frames if frame["subtype"] != "urllib3")

    outbound_frames = []
    for request, response in pairwise(without_urllib3):
        if (
            request["type"] != "outbound_http_request"
            or response["type"] != "outbound_http_response"
        ):
            # We don't have a (request, response) pair, so move ahead.
            # Usually this is (response, request), but could also be
            # (request, request) or (response, response) if the trace
            # failed to record as expected.
            continue  # pragma: no cover

        response = update_response_json_body(response)
        outbound_frame: OutboundFrame = {"request": request, "response": response}
        outbound_frames.append(outbound_frame)

    return outbound_frames



from __future__ import annotations

import types

from ..serialize import decode_body, decode_header_value


def process_urllib(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    request = frame.f_locals["req"]
    full_url = request.full_url
    method = request.get_method()
    method_and_full_url = f"{method} {full_url}"
    if event == "call":
        request_headers = {
            key: decode_header_value(value) for key, value in request.header_items()
        }

        return {
            "body": decode_body(request.data, request_headers),
            "headers": request_headers,
            "method": method,
            "method_and_full_url": method_and_full_url,
            "url": full_url,
        }

    elif event == "return":  # pragma: no branch
        if arg is None:
            # Probably an exception
            headers = None
            status_code = None
        else:
            response = frame.f_locals["r"]
            headers = dict(response.headers)
            status_code = response.status

        return {
            "body": None,
            "headers": headers,
            "method": method,
            "method_and_full_url": method_and_full_url,
            "status_code": status_code,
            "url": full_url,
        }

    elif event == "unwind":  # pragma: no cover
        return {
            "body": None,
            "headers": None,
            "method": method,
            "method_and_full_url": method_and_full_url,
            "status_code": None,
            "url": full_url,
        }


urllib = {
    "co_names": ("do_open",),
    "path_fragment": "urllib/request",
    "call_type": "outbound_http_request",
    "return_type": "outbound_http_response",
    "subtype": "urllib",
    "process": process_urllib,
}



from __future__ import annotations

import inspect
import logging
import platform
import sys
import threading
import time
import types
from collections import defaultdict
from collections.abc import Mapping
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Protocol, TypeVar, overload

import httpx
import msgpack
import ulid

from kolo.threads import get_thread_id

from .config import CONFIG_KEYS_TO_OMIT_FROM_SAVED_TRACE, load_config
from .db import save_trace_in_sqlite, setup_db
from .filters.attrs import attrs_filter
from .filters.core import (
    FrameFilter,
    FrameProcessor,
    exec_filter,
    frozen_filter,
    get_ignore_frames,
    get_include_frames,
    library_filter,
)
from .filters.kolo import kolo_filter
from .filters.pypy import pypy_filter
from .filters.pytest import pytest_generated_filter
from .git import COMMIT_SHA
from .plugins import PluginProcessor, load_plugin_data
from .serialize import (
    UserCodeCallSite,
    dump_msgpack,
    dump_msgpack_lightweight_repr,
    frame_path,
    monkeypatch_queryset_repr,
    user_code_call_site,
)
from .upload import upload_to_dashboard
from .utils import extract_http_trace_name
from .version import __version__

logger = logging.getLogger("kolo")


class KoloLocals(threading.local):
    def __init__(self):
        self.call_frames = []
        self._frame_ids = {}


INCLUDE_FRAMES_WARNING = """\
Unexpected exception in include_frames: %s
    co_filename: %s
    co_name: %s
    event: %s
    frame locals: %s
"""
IGNORE_FRAMES_WARNING = """\
Unexpected exception in ignore_frames: %s
    co_filename: %s
    co_name: %s
    event: %s
    frame locals: %s
"""
DEFAULT_INCLUDE_FRAMES_WARNING = """\
Unexpected exception in default_include_frames: %s
    co_filename: %s
    co_name: %s
    event: %s
    frame locals: %s
"""
DEFAULT_IGNORE_FRAMES_WARNING = """\
Unexpected exception in default_ignore_frames: %s
    co_filename: %s
    co_name: %s
    event: %s
    frame locals: %s
"""
PROCESS_FRAME_WARNING = """\
Unexpected exception in KoloProfiler.process_frame
    co_filename: %s
    co_name: %s
    event: %s
    frame locals: %s
"""


class KoloProfiler:
    """
    Collect runtime information about code to view in VSCode.

    include_frames can be passed to enable profiling of standard library
    or third party code.

    ignore_frames can also be passed to disable profiling of a user's
    own code.

    The list should contain fragments of the path to the relevant files.
    For example, to include profiling for the json module the include_frames
    could look like ["/json/"].

    The list may also contain frame filters. A frame filter is a function
    (or other callable) that takes the same arguments as the profilefunc
    passed to sys.setprofile and returns a boolean representing whether
    to allow or block the frame.

    include_frames takes precedence over ignore_frames. A frame that
    matches an entry in each list will be profiled.

    Rough order of operations in profiler (ideally kolo could natively make such a list!)
    1. __call__ is the sys registered profile callback which then calls calls push_frame_data
    2. push_frame_data adds to self.frames_by_thread
    3. save_trace_in_thread
    4. KoloProfiler.save
    5. build_trace
    """

    def __init__(
        self,
        db_path: Path,
        config=None,
        one_trace_per_test=False,
        *,
        source,
        name: Optional[str] = None,
    ) -> None:
        self.db_path = db_path
        self.source = source
        self.one_trace_per_test = one_trace_per_test
        trace_id = ulid.new()
        self.trace_id = f"trc_{trace_id}"
        self.trace_name = name
        self.start_test_indices: Dict[str, int] = {}
        self.config = config if config is not None else {}
        self.include_frames = get_include_frames(config)
        self.ignore_frames = get_ignore_frames(config)

        self.default_include_frames: Dict[str, List[FrameProcessor]] = {}
        for plugin_data in load_plugin_data(self.config):
            processor = PluginProcessor(plugin_data, self.config)
            for co_name in plugin_data["co_names"]:
                self.default_include_frames.setdefault(co_name, []).append(processor)

        self.default_ignore_frames: List[FrameFilter] = [
            library_filter,
            frozen_filter,
            pypy_filter,
            kolo_filter,
            exec_filter,
            attrs_filter,
            pytest_generated_filter,
        ]
        self.thread_locals = KoloLocals()
        self.timestamp = time.time()
        self.rust_profiler = None
        self.omit_return_locals = self.config.get("omit_return_locals", False)

        # Key is the thread id, value is the native python thread object
        self.threads: Dict[str, threading.Thread] = {}

        # Key is the thread id, value is a list of frames for that thread
        self.frames_by_thread: Dict[str, List[bytes]] = defaultdict(list)

        self.current_thread_id = get_thread_id(threading.current_thread())
        if self.config.get("lightweight_repr", False):
            self.dump_msgpack = dump_msgpack_lightweight_repr
        else:
            self.dump_msgpack = dump_msgpack

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> None:
        if event in ["c_call", "c_return", "c_exception"]:
            return

        co_name = frame.f_code.co_name

        frames = []
        frame_types = []

        try:
            # Execute only the filters listening for this co_name
            for processor in self.default_include_frames.get(co_name, ()):
                try:
                    if processor(frame, event, arg):
                        frame_data = processor.process(
                            frame, event, arg, self.thread_locals.call_frames
                        )
                        if frame_data:  # pragma: no branch
                            data = self.dump_msgpack(frame_data)
                            frames.append(data)
                            frame_types.append(frame_data["type"])
                except Exception as e:
                    logger.warning(
                        DEFAULT_INCLUDE_FRAMES_WARNING,
                        processor,
                        frame.f_code.co_filename,
                        frame.f_code.co_name,
                        event,
                        frame.f_locals,
                        exc_info=e,
                    )
                    continue

            for frame_filter in self.include_frames:
                try:
                    if frame_filter(frame, event, arg):
                        frames.append(self.process_frame(frame, event, arg))
                        frame_types.append("frame")
                        return
                except Exception as e:
                    logger.warning(
                        INCLUDE_FRAMES_WARNING,
                        frame_filter,
                        frame.f_code.co_filename,
                        frame.f_code.co_name,
                        event,
                        frame.f_locals,
                        exc_info=e,
                    )
                    continue

            for frame_filter in self.default_ignore_frames:
                try:
                    if frame_filter(frame, event, arg):
                        return
                except Exception as e:
                    logger.warning(
                        DEFAULT_IGNORE_FRAMES_WARNING,
                        frame_filter,
                        frame.f_code.co_filename,
                        frame.f_code.co_name,
                        event,
                        frame.f_locals,
                        exc_info=e,
                    )
                    continue

            for frame_filter in self.ignore_frames:
                try:
                    if frame_filter(frame, event, arg):
                        return
                except Exception as e:
                    logger.warning(
                        IGNORE_FRAMES_WARNING,
                        frame_filter,
                        frame.f_code.co_filename,
                        frame.f_code.co_name,
                        event,
                        frame.f_locals,
                        exc_info=e,
                    )
                    continue

            try:
                frames.append(self.process_frame(frame, event, arg))
                frame_types.append("frame")
            except Exception as e:
                logger.warning(
                    PROCESS_FRAME_WARNING,
                    frame.f_code.co_filename,
                    frame.f_code.co_name,
                    event,
                    frame.f_locals,
                    exc_info=e,
                )

        finally:
            if not frames:
                return

            if event == "return":
                frames.reverse()
                frame_types.reverse()

            if self.one_trace_per_test:  # pragma: no branch
                for index, frame_type in enumerate(frame_types):  # pragma: no cover
                    if frame_type == "start_test":
                        before, frames = frames[:index], frames[index:]

                        self.push_frame_data(before)

                        self.start_test()

                    elif frame_type == "end_test":
                        before, frames = frames[: index + 1], frames[index + 1 :]

                        self.push_frame_data(before)

                        self.end_test()

            self.push_frame_data(frames)

    def push_frame_data(self, data):
        current_thread = threading.current_thread()

        thread_id = get_thread_id(current_thread)
        if thread_id not in self.threads:
            self.threads[thread_id] = current_thread

        if thread_id not in self.frames_by_thread:
            self.frames_by_thread[thread_id] = []

        self.frames_by_thread[thread_id].extend(data)

    def start_test(self):
        self.trace_id = f"trc_{ulid.new()}"
        self.start_test_indices = {
            thread_id: len(frames)
            for thread_id, frames in self.frames_by_thread.items()
        }

    def end_test(self):
        frames_by_thread = {
            thread_id: frames[self.start_test_indices.get(thread_id, 0) :]
            for thread_id, frames in self.frames_by_thread.items()
        }
        self.save(frames_by_thread=frames_by_thread)

    def __enter__(self) -> None:
        if self.config.get("use_rust", True):
            try:
                from ._kolo import register_profiler
            except ImportError:
                # Useful for PyPY, which doesn't do rust
                sys.setprofile(self)
                threading.setprofile(self)
            else:
                register_profiler(self)
        else:
            sys.setprofile(self)
            threading.setprofile(self)

    def __exit__(self, *exc) -> None:
        sys.setprofile(None)
        threading.setprofile(None)

    def build_trace(self, frames_by_thread=None):
        # frames_by_thread is only passed when called by end_test.

        if self.rust_profiler:
            return self.rust_profiler.build_trace()

        config = {
            k: v
            for k, v in self.config.items()
            if k not in CONFIG_KEYS_TO_OMIT_FROM_SAVED_TRACE
        }
        # Config to ALWAYS include, even if the user user didn't change the default
        config["use_monitoring"] = False
        config["use_rust"] = False

        timestamp = self.timestamp
        data = {
            "command_line_args": sys.argv,
            "current_commit_sha": COMMIT_SHA,
            "current_thread_id": self.current_thread_id,
            "meta": {
                "version": __version__,
                "source": self.source,
                "environment": {
                    "py_version": platform.python_version(),
                    "py_version_full": sys.version,
                    "platform": platform.platform(),
                    "system": platform.system(),
                    "machine": platform.machine(),
                    "processor": platform.processor(),
                },
                "config": config,
            },
            "timestamp": timestamp,
            "trace_id": self.trace_id,
            "trace_name": self.trace_name,
        }

        if frames_by_thread is None:
            frames_by_thread = self.frames_by_thread

        # From kolo v2.35.0 onwards, frames_of_interest and frames are always empty
        data["frames_of_interest"] = []
        data["frames"] = {}

        data["threads"] = {}
        for thread_id, thread in self.threads.items():
            data["threads"][thread_id] = {
                "name": thread.name,
                "ident": getattr(thread, "ident", None),
                "native_id": getattr(thread, "native_id", None),
                "daemon": thread.daemon,
                "is_alive": thread.is_alive(),
            }

            if thread_id in frames_by_thread:
                data["threads"][thread_id]["frames"] = [
                    msgpack.unpackb(f, strict_map_key=False)
                    for f in frames_by_thread[thread_id]
                ]

        return self.dump_msgpack(data)

    def save(self, frames_by_thread=None) -> None:
        """
        frames_by_thread is only passed when called from end_test,
        because end_test cuts off some frames and saves directly.
        """

        if self.rust_profiler:
            self.rust_profiler.save()
            return

        if self.trace_name is None:
            self._set_http_trace_name(frames_by_thread)

        serialized_data = self.build_trace(frames_by_thread=frames_by_thread)
        timeout = self.config.get("sqlite_busy_timeout", 60)
        save_trace_in_sqlite(
            self.db_path, self.trace_id, msgpack=serialized_data, timeout=timeout
        )

    def _set_http_trace_name(self, frames_by_thread=None):
        """
        Extract HTTP request/response information from frames to set the trace name.
        """
        if frames_by_thread is None:
            frames_by_thread = self.frames_by_thread

        trace_name = extract_http_trace_name(frames_by_thread, self.current_thread_id)
        if trace_name:
            self.trace_name = trace_name

    def process_frame(self, frame: types.FrameType, event: str, arg: object) -> bytes:
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self.thread_locals._frame_ids[id(frame)] = frame_id
        elif event == "return":  # pragma: no branch
            frame_id = self.thread_locals._frame_ids[id(frame)]

        user_code_call_site_ = user_code_call_site(
            self.thread_locals.call_frames, frame_id
        )

        if event == "call":
            self.thread_locals.call_frames.append((frame, frame_id))
        elif event == "return":  # pragma: no branch
            self.thread_locals.call_frames.pop()

        if self.omit_return_locals and event == "return":
            frame_locals = None
        else:
            frame_locals = {
                k: v for k, v in frame.f_locals.items() if k != "__builtins__"
            }

        frame_data = {
            "path": frame_path(frame),
            "co_name": frame.f_code.co_name,
            "qualname": get_qualname(frame),
            "event": event,
            "frame_id": frame_id,
            "arg": arg,
            "locals": frame_locals,
            "timestamp": time.time(),
            "type": "frame",
            "user_code_call_site": user_code_call_site_,
        }
        return self.dump_msgpack(frame_data)


def get_qualname(frame: types.FrameType) -> str | None:
    try:
        qualname = frame.f_code.co_qualname  # type: ignore[attr-defined]
    except AttributeError:
        pass
    else:
        module = frame.f_globals.get("__name__", "<unknown>")
        return f"{module}.{qualname}"

    co_name = frame.f_code.co_name
    if co_name == "<module>":  # pragma: no cover
        module = frame.f_globals.get("__name__", "<unknown>")
        return f"{module}.<module>"

    try:
        outer_frame = frame.f_back
        assert outer_frame
        try:
            function = outer_frame.f_locals[co_name]
        except KeyError:
            try:
                self = frame.f_locals["self"]
            except KeyError:
                cls = frame.f_locals.get("cls")
                if isinstance(cls, type):
                    function = inspect.getattr_static(cls, co_name)
                else:
                    try:
                        qualname = frame.f_locals["__qualname__"]
                    except KeyError:
                        function = frame.f_globals[co_name]
                    else:  # pragma: no cover
                        module = frame.f_globals.get("__name__", "<unknown>")
                        return f"{module}.{qualname}"
            else:
                function = inspect.getattr_static(self, co_name)
                if isinstance(function, property):
                    function = function.fget

        return f"{function.__module__}.{function.__qualname__}"
    except Exception:
        return None


def save_trace_in_thread(profiler):
    if platform.machine() == "wasm32":
        profiler.save()
    else:
        name = "kolo-save_request_in_db"
        threading.Thread(target=profiler.save, name=name).start()


def upload_trace_in_thread(profiler, upload_token):
    def upload():
        trace = profiler.build_trace()
        try:
            response = upload_to_dashboard(trace, upload_token)
            response.raise_for_status()
        except httpx.HTTPError:
            logger.exception("Failed to upload trace to Kolo dashboard.")

    if platform.machine() == "wasm32":
        upload()
    else:
        name = "kolo-upload_to_dashboard"
        threading.Thread(target=upload, name=name).start()


class Enabled:
    def __init__(
        self,
        config: Mapping[str, Any] | None,
        source: str,
        one_trace_per_test: bool,
        save_in_thread: bool,
        upload_token: Optional[str],
        db_path: Optional[Path] = None,
        name: Optional[str] = None,
    ):
        if config is None:
            config = {}
        self.config = load_config(config)
        self._profiler: KoloProfiler | None = None
        self._monitor = None
        self.source = source
        self.one_trace_per_test = one_trace_per_test
        self.save_in_thread = save_in_thread
        self.upload_token = upload_token
        self.db_path = db_path
        self.name = name

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner

    def __enter__(self) -> None:
        use_monitoring = self.config.get("use_monitoring", False)

        if use_monitoring and sys.version_info >= (3, 12):
            if sys.monitoring.get_tool(sys.monitoring.PROFILER_ID):  # type: ignore[attr-defined]
                return
        elif sys.getprofile():
            return

        if sys.version_info < (3, 12) or not use_monitoring:
            try:
                thread_profiler = threading.getprofile()  # type: ignore[attr-defined]
            except AttributeError:
                thread_profiler = threading._profile_hook
            if thread_profiler:
                return

        # self.db_path is typically not set,
        # but we make use of it in tests.
        if self.db_path is None:
            db_path = setup_db()
        else:
            db_path = self.db_path

        monkeypatch_queryset_repr()

        if sys.version_info >= (3, 12) and use_monitoring:
            from .monitoring import activate_monitoring

            monitor = self.register_monitor(db_path)
            activate_monitoring(monitor)
            self._monitor = monitor
        else:
            self._profiler = KoloProfiler(
                db_path,
                config=self.config,
                source=self.source,
                one_trace_per_test=self.one_trace_per_test,
                name=self.name,
            )
            self._profiler.__enter__()

    def register_monitor(self, db_path):
        use_rust = self.config.get("use_rust", True)
        if use_rust:
            from ._kolo import register_monitor

            return register_monitor(
                str(db_path),
                config=self.config,
                source=self.source,
                one_trace_per_test=self.one_trace_per_test,
                name=self.name,
            )

        from .monitoring import KoloMonitor  # type: ignore[attr-defined]

        return KoloMonitor(
            db_path,
            config=self.config,
            source=self.source,
            one_trace_per_test=self.one_trace_per_test,
            name=self.name,
        )

    def save_trace_profiler(self):
        assert self._profiler is not None

        if self.one_trace_per_test:
            return

        if not self.save_in_thread:
            self._profiler.save()
            return

        if self.upload_token:
            upload_trace_in_thread(self._profiler, self.upload_token)
        else:
            save_trace_in_thread(self._profiler)

    def save_trace_monitor(self):
        assert self._monitor is not None

        if self.one_trace_per_test:
            return

        if not self.save_in_thread:
            self._monitor.save()
            return

        if self.upload_token:
            upload_trace_in_thread(self._monitor, self.upload_token)
        else:
            save_trace_in_thread(self._monitor)

    def __exit__(self, *exc) -> None:
        if self._profiler is not None:
            self._profiler.__exit__(*exc)
            self.save_trace_profiler()
            self._profiler = None

        if self._monitor is not None:
            from .monitoring import disable_monitoring  # type: ignore[attr-defined]

            disable_monitoring(self._monitor)
            self.save_trace_monitor()
            self._monitor = None


F = TypeVar("F", bound=Callable[..., Any])


class CallableContextManager(Protocol):
    def __call__(self, func: F) -> F: ...  # pragma: no cover

    def __enter__(self) -> None: ...  # pragma: no cover

    def __exit__(self, *exc) -> None: ...  # pragma: no cover


@overload
def enable(_func: F) -> F:
    """Stub"""


@overload
def enable(
    config: Mapping[str, Any] | None = None,
    name: Optional[str] = None,
    source: str = "kolo.enable",
    _one_trace_per_test: bool = False,
    _save_in_thread: bool = False,
    _upload_token: Optional[str] = None,
    _db_path: Optional[Path] = None,
) -> CallableContextManager:
    """Stub"""


def enable(
    config=None,
    *,
    name=None,
    source="kolo.enable",
    _one_trace_per_test=False,
    _save_in_thread=False,
    _upload_token=None,
    _db_path=None,
):
    if config is None or isinstance(config, Mapping):
        function = None
    else:
        # Treat as a decorator called on a function
        function = config
        config = None

    enabled = Enabled(
        config=config,
        source=source,
        name=name,
        one_trace_per_test=_one_trace_per_test,
        save_in_thread=_save_in_thread,
        upload_token=_upload_token,
        db_path=_db_path,
    )

    if function is None:
        return enabled
    return enabled(function)


enabled = enable



from __future__ import annotations

import types

import ulid


def pytest_generated_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore pytest generated code

    When using the `-k` or `-m` command line argument, pytest compiles
    code with a custom filename:

        <pytest match expression>

    https://github.com/pytest-dev/pytest/blob/9454fc38d3636b79ee657d6cacf7477eb8acee52/src/_pytest/mark/expression.py#L208

    Since this is library code, we want to filter it out.
    """
    return frame.f_code.co_filename == "<pytest match expression>"


def build_context(config):
    return {"frame_ids": {}}


def process_pytest(frame, event, arg, context):  # pragma: no cover
    frame_ids = context["frame_ids"]
    co_name = frame.f_code.co_name
    # location is a (filename, lineno, testname) tuple
    # https://docs.pytest.org/en/7.1.x/reference/reference.html#pytest.hookspec.pytest_runtest_logstart
    location = frame.f_locals["location"]
    if co_name == "pytest_runtest_logstart":
        frame_id = f"frm_{ulid.new()}"
        frame_ids[id(location)] = frame_id
    else:
        frame_id = frame_ids[id(location)]

    _filename, _lineno, test = location
    test_class, _sep, test_name = test.rpartition(".")
    return {
        "frame_id": frame_id,
        "test_name": test_name,
        "test_class": test_class if test_class else None,
    }


def call_pytest(frame, event, arg, context):
    return (  # pragma: no cover
        event == "call" and frame.f_code.co_name == "pytest_runtest_logstart"
    ) or (event == "return" and frame.f_code.co_name == "pytest_runtest_logfinish")


pytest = {
    "co_names": ("pytest_runtest_logstart", "pytest_runtest_logfinish"),
    "path_fragment": "kolo/pytest_plugin.py",
    "call": call_pytest,
    "call_type": "start_test",
    "return_type": "end_test",
    "process": process_pytest,
    "build_context": build_context,
}



from __future__ import annotations

import types


def call_celery(frame, event, arg, context):
    return "sentry_sdk" not in frame.f_code.co_filename


def process_celery(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    frame_locals = frame.f_locals
    task_name = frame_locals["self"].name
    task_args = frame_locals["args"]
    task_kwargs = frame_locals["kwargs"]
    return {
        "name": task_name,
        "args": task_args,
        "kwargs": task_kwargs,
    }


celery = {
    "co_names": ("apply_async",),
    "path_fragment": "celery/app/task.py",
    "call": call_celery,
    "call_type": "background_job",
    "return_type": "background_job_end",
    "subtype": "celery",
    "process": process_celery,
}



from typing import Dict, List

from .cli_mcp_shared import get_compact_trace as aget_compact_trace
from .cli_mcp_shared import get_compact_traces, get_formatted_traces, get_node_data
from .db import get_db_path, load_trace_from_db, load_trace_with_size_from_db, setup_db
from .mcp_plumbing import MCPServer

# Create a global MCP server instance
mcp = MCPServer("kolo")


@mcp.tool()
async def get_compact_trace(trace_id: str, include_returns: bool = False) -> str:
    """Get a compact representation of a specific trace.

    This is useful when you need to understand what happened in a specific trace,
    identified by its trace_id. The compact representation shows the sequence of
    function calls and their outcomes.

    Args:
        trace_id: ID of the trace to get
        include_returns: Whether to include return values (warning: can be verbose)

    Example:
        >>> get_compact_trace("trc_1234567890")
        "main() -> process_data() -> validate_input() -> ..."
    """
    db_path = get_db_path()
    _, timestamp_str, size, trace_data = load_trace_with_size_from_db(db_path, trace_id)
    return await aget_compact_trace(
        trace_id=trace_id,
        trace_data=trace_data,
        timestamp_str=timestamp_str,
        size=size,
        include_returns=include_returns,
    )


@mcp.tool()
async def get_trace_node(trace_id: str, node_index: int) -> Dict:
    """Get detailed information about a specific node in a trace.

    This is useful when you need to deeply understand what happened at a specific
    point in the execution, like examining function arguments, local variables,
    or the exact line of code being executed.

    Args:
        trace_id: ID of the trace containing the node
        node_index: Index of the node in the trace tree

    Example:
        >>> get_trace_node("trc_123", 5)
        {"type": "call", "function": "validate_input", "args": {"data": {...}}, ...}
    """
    db_path = get_db_path()
    trace_data, _ = load_trace_from_db(db_path, trace_id)
    return await get_node_data(trace_id, node_index, trace_data)


@mcp.tool()
async def get_pinned_traces(include_returns: bool = False) -> List[str]:
    """Get compact representations of all pinned traces.

    This is useful for understanding the key execution paths that have been marked
    as important. Pinned traces often represent core functionality or critical
    paths through the codebase.

    Args:
        include_returns: Whether to include return values (warning: can be verbose)

    Example:
        >>> get_pinned_traces()
        [
            "main() -> process_data() -> ...",
            "test() -> validate() -> ..."
        ]
    """
    db_path = get_db_path()
    traces = await get_compact_traces(db_path, pinned=True, returns=include_returns)
    return [compact for _, compact in traces]


@mcp.tool()
async def get_recent_compact_traces(
    count: int = 5, include_returns: bool = False
) -> List[str]:
    """
    Get the compact representation of the most recent traces.

    The traces are detailed call graphs of program execution. It's as if a debugger was active
    for the last 5 requests/program executions and every function call and return was logged.
    It includes extremely useful data about program execution flow, function arguments, return
    values, sql queries, and more.

    Use this tool to get unstuck and to understand how the just executed code actually behaved
    in practice. Use it to check assumptions or to answer the question "why" something didn't work.

    This tool call is cheap and fast, when in doubt, make it to gather extra info.

    Examples:
    - A 500 error happened. The kolo trace includes the full stack trace revealing why the error happened
    - Everything seems fine, but the user is reporting unexpected behaviour. The kolo trace reveals a codepath you
    didn't expect and need to account for.
    - You're debugging a complex issue and need to understand the flow of the program. The kolo trace shows you

    Warning: Setting include_returns=True with a large count can return a lot of data.

    Args:
        count: Number of recent traces to get (default: 5)
        include_returns: Whether to include return values (warning: can be verbose)

    Example:
        >>> get_recent_compact_traces(2)
        [
            "main() -> process_data() -> ...",
            "test() -> validate() -> ..."
        ]
    """
    db_path = get_db_path()
    traces = await get_compact_traces(db_path, recent=count, returns=include_returns)
    return [compact for _, compact in traces]


@mcp.tool()
def list_traces(count: int = 30) -> List[str]:
    """Concisely list available traces, ordered by most recent first.

    This is useful for getting an overview of what traces are available and
    finding specific traces you might want to analyze further.
    It's also useful to see recent activity. Traces may be named.
    Call this tool *often* to orient yourself and see where useful context may lie.
    After using this tool, call get_compact_trace to get more details about a specific trace.

    Args:
        count: How many traces to list (default: 30)

    Example:
        >>> list_traces(5)
        [
            "200 POST /buttons/poll_action/ (20h ago,   1.3 MB, trc_01JPMVDD0TRMW30JDAWR8HR4ZR)",
            "200 POST /poll/command/poll/ (20h ago, 321.1 KB, trc_01JPMVDAE8TA7N06JHJ809K2PE)",
            "200 POST /buttons/poll_action/vote/ (21h ago, 384.9 KB, trc_01JPMPK97J568NH2JZP5J2P02F)",
            "trc_01JPMKX2BEYZT7VM8A3FWC6DVK (22h ago, 346.8 KB)",
            "trc_01JPMKTVE3X7WRRE06Y2BDWXZ5 (22h ago, 346.8 KB)"
        ]
    """
    db_path = setup_db()

    # get_formatted_traces returns an iterator, not an awaitable
    formatted_traces = get_formatted_traces(db_path, count=count)
    return list(formatted_traces)



from __future__ import annotations

import gzip
import json
import logging
import os
import types
from datetime import date, datetime, time
from decimal import Decimal
from email.message import Message
from email.utils import collapse_rfc2231_value
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, TypedDict, TypeVar

try:
    from django.db.models import QuerySet
except ImportError:
    QuerySet = None  # type: ignore

import msgpack

logger = logging.getLogger("kolo")


if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse, StreamingHttpResponse


class UserCodeCallSite(TypedDict):
    line_number: int
    call_frame_id: str


def user_code_call_site(
    call_frames: List[Tuple[types.FrameType, str]],
    frame_id: str,
) -> UserCodeCallSite | None:
    count = len(call_frames)
    if count == 0:
        return None

    call_frame, call_frame_id = call_frames[-1]

    # If `frame_id` is set in a `return` frame, we want to skip it
    if call_frame_id == frame_id:
        if count == 1:
            return None
        call_frame, call_frame_id = call_frames[-2]

    return {
        "call_frame_id": call_frame_id,
        "line_number": call_frame.f_lineno,
    }


Local = TypeVar("Local")


SERIALIZE_PATH = os.path.normpath("kolo/serialize.py")


# TODO: Make these threadlocals when we support multithreading
QUERYSET_PATCHED = False
IN_KOLO_PROFILER = False


def monkeypatch_queryset_repr():
    global QUERYSET_PATCHED
    if QUERYSET_PATCHED:
        return

    try:
        from django.db.models import QuerySet
    except ImportError:  # pragma: no cover
        QUERYSET_PATCHED = True
        return

    old_repr = QuerySet.__repr__

    def new_repr(queryset):
        if getattr(queryset, "_result_cache", None) is None and IN_KOLO_PROFILER:
            return f"Unevaluated queryset for: {queryset.model}"
        return old_repr(queryset)

    QuerySet.__repr__ = new_repr  # type: ignore
    QUERYSET_PATCHED = True


def msgpack_encode_hook(obj):
    try:
        if isinstance(obj, tuple):
            return msgpack.ExtType(6, _dump_msgpack(list(obj)))

        if isinstance(obj, set):
            return msgpack.ExtType(7, _dump_msgpack(list(obj)))

        if isinstance(obj, frozenset):
            return msgpack.ExtType(8, _dump_msgpack(list(obj)))

        if isinstance(obj, dict):
            return dict(obj)

        if isinstance(obj, datetime):
            return msgpack.ExtType(1, obj.isoformat().encode("utf-8"))

        if isinstance(obj, date):
            return msgpack.ExtType(2, obj.isoformat().encode("utf-8"))

        if isinstance(obj, time):
            return msgpack.ExtType(3, obj.isoformat().encode("utf-8"))

        if isinstance(obj, int):
            num_bytes = (obj.bit_length() + 7) // 8 + 1
            return msgpack.ExtType(4, obj.to_bytes(num_bytes, "big", signed=True))

        if isinstance(obj, Decimal):
            return msgpack.ExtType(5, str(obj).encode("utf-8"))

        if (
            QuerySet is not None
            and isinstance(obj, QuerySet)
            and obj._result_cache is None
        ):
            data = f"Unevaluated queryset for: {obj.model}".encode("utf-8")
            return msgpack.ExtType(0, data)

        return msgpack.ExtType(0, repr(obj).encode("utf-8"))
    except Exception:
        path = f"{obj.__module__}.{type(obj).__qualname__}"
        return msgpack.ExtType(127, path.encode("utf-8"))


def msgpack_decode_hook(code, data):
    if code == 0:
        return data.decode("utf-8")
    if code == 1:
        return datetime.fromisoformat(data.decode("utf-8"))
    if code == 2:
        return date.fromisoformat(data.decode("utf-8"))
    if code == 3:
        return time.fromisoformat(data.decode("utf-8"))
    if code == 4:
        return int.from_bytes(data, "big", signed=True)
    if code == 5:
        return Decimal(data.decode("utf-8"))
    if code == 6:
        return tuple(load_msgpack(data))
    if code == 7:
        return set(load_msgpack(data))
    if code == 8:
        return frozenset(load_msgpack(data))
    if code == 127:
        data = data.decode("utf-8")
        return f"KoloSerializationError: {data}"
    raise ValueError(f"Unknown msgpack extension. Code: {code}")


def _dump_msgpack(data):
    return msgpack.packb(data, default=msgpack_encode_hook, strict_types=True)


def dump_msgpack(data):
    global IN_KOLO_PROFILER
    IN_KOLO_PROFILER = True
    try:
        return _dump_msgpack(data)
    finally:
        IN_KOLO_PROFILER = False


def msgpack_encode_hook_lightweight_repr(obj):
    if isinstance(obj, tuple):
        return msgpack.ExtType(6, dump_msgpack_lightweight_repr(list(obj)))

    if isinstance(obj, set):
        return msgpack.ExtType(7, dump_msgpack_lightweight_repr(list(obj)))

    if isinstance(obj, frozenset):
        return msgpack.ExtType(8, dump_msgpack_lightweight_repr(list(obj)))

    if isinstance(obj, dict):
        return dict(obj)

    if isinstance(obj, datetime):
        return msgpack.ExtType(1, obj.isoformat().encode("utf-8"))

    if isinstance(obj, date):
        return msgpack.ExtType(2, obj.isoformat().encode("utf-8"))

    if isinstance(obj, time):
        return msgpack.ExtType(3, obj.isoformat().encode("utf-8"))

    if isinstance(obj, int):
        num_bytes = (obj.bit_length() + 7) // 8 + 1
        return msgpack.ExtType(4, obj.to_bytes(num_bytes, "big", signed=True))

    if isinstance(obj, Decimal):
        return msgpack.ExtType(5, str(obj).encode("utf-8"))

    return msgpack.ExtType(
        0, f"<{obj.__class__.__qualname__} object {id(obj)}>".encode("utf-8")
    )


def dump_msgpack_lightweight_repr(data):
    return msgpack.packb(
        data, default=msgpack_encode_hook_lightweight_repr, strict_types=True
    )


def load_msgpack(data):
    return msgpack.unpackb(data, strict_map_key=False, ext_hook=msgpack_decode_hook)


def decode_header_value(bytes_or_str: bytes | str) -> str:
    """
    Convert a bytes header value to text.

    Valid header values are expected to be ascii in modern times, but
    ISO-8859-1 (latin1) has historically been allowed.

    https://datatracker.ietf.org/doc/html/rfc7230#section-3.2.4
    """
    if isinstance(bytes_or_str, bytes):
        return bytes_or_str.decode("latin1")
    return bytes_or_str


def frame_path(frame: types.FrameType) -> str:
    path = frame.f_code.co_filename
    try:
        relative_path = os.path.relpath(path)
    except ValueError:
        relative_path = path
    return f"{relative_path}:{frame.f_lineno}"


def decode_body(body: Any, request_headers: Dict[str, str]) -> Any:
    """Convert a request body into a json-serializable form."""
    if isinstance(body, bytes):
        content_type = request_headers.get("Content-Type", "")
        m = Message()
        m["content-type"] = content_type
        charset = collapse_rfc2231_value(m.get_param("charset", "utf-8"))
        try:
            return body.decode(charset)
        except UnicodeDecodeError:
            return "<Binary request body>"
    return body


def get_content(response: HttpResponse | StreamingHttpResponse) -> str:
    if response.streaming:
        return "<Streaming Response>"

    if TYPE_CHECKING:
        assert isinstance(response, HttpResponse)
    content_encoding = response.get("Content-Encoding")
    if content_encoding == "gzip":
        content = gzip.decompress(response.content)
    else:
        content = response.content
    try:
        return content.decode(response.charset)
    except UnicodeDecodeError:
        return f"<Response with invalid charset ({response.charset})>"


def get_request_body(request: "HttpRequest") -> str:
    from django.http.request import RawPostDataException

    try:
        return request.body.decode("utf-8")
    except UnicodeDecodeError:  # pragma: no cover
        return "<Binary request body>"
    except RawPostDataException:
        return "<Request data already read>"



from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator, List, Optional, Tuple, Union

import httpx

from .db import (
    get_pinned_traces,
    list_traces_with_data_from_db,
    load_trace_with_size_from_db,
)
from .serialize import load_msgpack
from .utils import pretty_byte_size


def relative_time(timestamp: datetime) -> str:
    now = datetime.now(timezone.utc)

    delta = now - timestamp
    if delta.days > 7:
        relative_time = timestamp.strftime("%Y-%m-%d")
    elif delta.days > 0:
        relative_time = f"{delta.days}d ago"
    elif delta.seconds > 3600:
        relative_time = f"{delta.seconds // 3600}h ago"
    elif delta.seconds > 60:
        relative_time = f"{delta.seconds // 60}m ago"
    else:
        relative_time = f"{delta.seconds}s ago"

    return relative_time


def format_trace_for_display(
    trace_id: str,
    timestamp: datetime,
    size: int,
    trace_name: Optional[str],
    now: Optional[datetime] = None,
) -> str:
    """Format a trace for display in CLI or MCP."""
    if now is None:
        now = datetime.now(timezone.utc)

    rel_time = relative_time(timestamp)

    size_str = pretty_byte_size(size)

    # Format output with trace_id at the end if we have a name
    if trace_name:
        return f"{trace_name} ({rel_time}, {size_str}, {trace_id})"
    return f"{trace_id} ({rel_time}, {size_str})"


def get_formatted_traces(
    db_path, count: int = 500, reverse: bool = False
) -> Iterator[str]:
    """Get formatted traces from the database, yielding one at a time."""
    now = datetime.now(timezone.utc)
    traces = list_traces_with_data_from_db(db_path, count=count, reverse=reverse)

    for trace_id, timestamp_str, size, msgpack_data in traces:
        timestamp = datetime.strptime(timestamp_str, "%Y-%m-%d %H:%M:%S.%f").replace(
            tzinfo=timezone.utc
        )

        # Extract trace name if available
        trace_name = None
        if msgpack_data:
            data = load_msgpack(msgpack_data)
            if isinstance(data, dict) and "trace_name" in data and data["trace_name"]:
                trace_name = data["trace_name"]

        yield format_trace_for_display(trace_id, timestamp, size, trace_name, now)


local_cf_worker_endpoint = "http://localhost:8787"
prod_cf_worker_endpoint = "https://worker.kolo.app"
worker_endpoint = prod_cf_worker_endpoint


async def get_compact_trace(
    trace_id: str,
    trace_data: bytes,
    timestamp_str: str,
    size: int,
    include_returns: bool = False,
) -> str:
    """Get compact representation of a trace from the worker."""

    compact_trace_text = await fetch_compact_trace(
        trace_id, trace_data, include_returns
    )

    timestamp = datetime.strptime(timestamp_str, "%Y-%m-%d %H:%M:%S.%f").replace(
        tzinfo=timezone.utc
    )

    prefix = f"{relative_time(timestamp)}, {pretty_byte_size(size)}"

    return prefix + "\n\n" + compact_trace_text


async def fetch_compact_trace(
    trace_id: str, trace_data: bytes, include_returns: bool = False
) -> str:
    """Get compact representation of a trace from the worker."""

    async with httpx.AsyncClient() as client:
        url = f"{worker_endpoint}/traces/{trace_id}/compact"
        if include_returns:
            url += "?returns=1"

        response = await client.post(
            url, headers={"Content-Type": "application/msgpack"}, content=trace_data
        )
        response.raise_for_status()
        return response.text


async def get_node_data(trace_id: str, node_index: int, trace_data: bytes) -> dict:
    """Get node data from the worker for a specific node index."""
    async with httpx.AsyncClient() as client:
        response = await client.post(
            f"{worker_endpoint}/traces/{trace_id}/tree/{node_index}",
            headers={"Content-Type": "application/msgpack"},
            content=trace_data,
        )
        response.raise_for_status()
        return response.json()


async def get_compact_traces(
    db_path: Path,
    trace_id: Union[str, None] = None,
    *,
    pinned: bool = False,
    returns: bool = False,
    recent: int = 0,
) -> List[Tuple[str, str]]:
    """Get compact representation of traces.

    Args:
        db_path: Path to the database
        trace_id: Specific trace ID to get, ignored if pinned=True or recent>0
        pinned: If True, get all pinned traces
        returns: Include return values in compact representation
        recent: If > 0, get the N most recent traces

    Returns:
        List of tuples (trace_id, compact_representation)

    Raises:
        TraceNotFoundError: If trace_id is not found in the database
        ValueError: If no valid selection criteria provided
    """
    if not any([pinned, trace_id, recent > 0]):
        raise ValueError("Either trace_id, --pinned, or --recent must be provided")
    results = []

    if pinned:
        for trace_id, timestamp_str, size, trace_data in get_pinned_traces(db_path):
            try:
                compact_repr = await get_compact_trace(
                    trace_id, trace_data, timestamp_str, size, returns
                )
                results.append((trace_id, compact_repr))
            except Exception as e:
                # For pinned traces, we want to continue even if one fails
                results.append((trace_id, f"Error: {e}"))
    elif recent > 0:
        for trace_id, timestamp_str, size, trace_data in list_traces_with_data_from_db(
            db_path, count=recent
        ):
            assert trace_id is not None
            try:
                compact_repr = await get_compact_trace(
                    trace_id, trace_data, timestamp_str, size, returns
                )
                results.append((trace_id, compact_repr))
            except Exception as e:
                # For multiple traces, we want to continue even if one fails
                results.append((trace_id, f"Error: {e}"))
    else:
        # Single trace - let errors propagate naturally
        assert trace_id is not None
        _, timestamp_str, size, trace_data = load_trace_with_size_from_db(
            db_path, trace_id
        )
        compact_repr = await get_compact_trace(
            trace_id, trace_data, timestamp_str, size, returns
        )
        results.append((trace_id, compact_repr))

    return results



import glob
import os
import subprocess
import sys

# We open VS Code source code through a URI scheme directly from the frontend.
# However, such a scheme doesn't seem to exist for PyCharm.
# This is the most reliable approach I managed to find for opening PyCharm files.


def get_pycharm_executable():
    """Get the PyCharm executable path based on the operating system."""
    if sys.platform.startswith("win"):
        # Windows: Check common installation directories
        common_paths = [
            r"C:\Program Files\JetBrains\PyCharm*\bin\pycharm64.exe",
            r"C:\Program Files (x86)\JetBrains\PyCharm*\bin\pycharm.exe",
        ]
        for path in common_paths:
            matches = sorted(glob.glob(path), reverse=True)
            if matches:
                return matches[0]
    elif sys.platform.startswith("darwin"):
        # macOS
        return "/Applications/PyCharm.app/Contents/MacOS/pycharm"
    else:
        # Linux
        return "pycharm"  # Assumes PyCharm is in PATH


def get_project_path(file_path):
    """
    Attempt to find the PyCharm project path by looking for .idea directory
    or a pyproject.toml file in parent directories.
    """
    current_dir = os.path.dirname(os.path.abspath(file_path))
    while current_dir != os.path.dirname(current_dir):  # Stop at root directory
        if os.path.exists(os.path.join(current_dir, ".idea")):
            return current_dir
        if os.path.exists(os.path.join(current_dir, "pyproject.toml")):
            return current_dir
        current_dir = os.path.dirname(current_dir)
    return None  # If no project directory found


class OpenInPyCharmError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def open_in_pycharm(file_path):
    pycharm_exe = get_pycharm_executable()
    if not pycharm_exe:
        raise OpenInPyCharmError("PyCharm executable not found.")

    project_path = get_project_path(file_path)
    if not project_path:
        # Unable to determine project path. Opening file directly.
        project_path = os.path.dirname(file_path)

    cmd = [pycharm_exe, project_path, file_path]

    try:
        subprocess.Popen(cmd)
    except Exception as e:
        OpenInPyCharmError(f"Error opening PyCharm: {e}")


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: python open_in_pycharm.py <file_path>")
        sys.exit(1)

    file_path = sys.argv[1]
    open_in_pycharm(file_path)



from __future__ import annotations

import time
import types
from contextlib import suppress

from ..serialize import get_content, get_request_body


def process_django(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    if event in ("call", "resume", "throw"):
        context["timestamp"] = time.time()
        request = frame.f_locals["request"]
        return {
            "body": get_request_body(request),
            "headers": dict(request.headers),
            "method": request.method,
            "path_info": request.path_info,
            # dict(request.POST) can give confusing results due to MultivalueDict
            "post_data": dict(request.POST),
            "query_params": dict(request.GET),
            "scheme": request.scheme,
            # Override the timestamp in PluginProcessor because consistency with
            # ms_duration is more important.
            "timestamp": context["timestamp"],
        }
    elif event in ("return", "yield", "unwind"):  # pragma: no branch
        timestamp = time.time()
        duration = timestamp - context["timestamp"]
        ms_duration = round(duration * 1000, 2)

        request = frame.f_locals["request"]
        match = request.resolver_match
        if match:  # match is None if this is a 404
            # TODO(later): We should record the urlconf used, if it's not ROOT_URLCONF
            # TODO(later): We should report on whether it was a `path` or `re_path` if we can;
            #  we might have to look in `tried` or go up a frame to find the pattern object?
            try:
                route_params = {
                    "captured": match.captured_kwargs,
                    "defaults": match.extra_kwargs,
                }
            except AttributeError:
                # Prior to Django 4.1, the match never holds a reference to the "default" arguments
                # etc; see the PR & corresponding ticket:
                # https://github.com/django/django/pull/15402
                # https://code.djangoproject.com/ticket/16406
                route_params = {
                    "captured": {},
                    "defaults": {},
                }
            url_pattern = {
                "namespace": match.namespace,
                "route": match.route,
                "route_params": route_params,
                "url_name": match.url_name,
                "view_qualname": match._func_path,
                "view_params": {"positional": match.args, "keyword": match.kwargs},
            }
        else:
            url_pattern = None

        if "response" not in frame.f_locals:
            return {
                "ms_duration": ms_duration,
                "status_code": None,
                "content": "",
                "headers": {},
                "url_pattern": url_pattern,
                # Override the timestamp in PluginProcessor because consistency with
                # ms_duration is more important.
                "timestamp": timestamp,
            }
        response = frame.f_locals["response"]
        return {
            "ms_duration": ms_duration,
            "status_code": response.status_code,
            "content": get_content(response),
            "headers": dict(response.items()),
            "url_pattern": url_pattern,
            # Override the timestamp in PluginProcessor because consistency with
            # ms_duration is more important.
            "timestamp": timestamp,
        }


def flatten(context):
    from django.template.context import BaseContext

    flat = {}
    for c in context.dicts:
        if isinstance(c, BaseContext):
            c = flatten(c)
        flat.update(c)
    return flat


def process_django_template(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    template_context = frame.f_locals["context"]
    with suppress(AttributeError):
        try:
            template_context = template_context.flatten()
        except ValueError:  # pragma: no cover
            # https://code.djangoproject.com/ticket/35417
            template_context = flatten(template_context)

    template_name = frame.f_locals["self"].name
    if isinstance(template_name, str):
        # `name` can be a Django `SafeString`. In this case, it would get
        # serialized as the `repr`, rather than the raw string we care about.
        # Appending "" converts back to a normal python `str`.
        template_name += ""

    return {
        "context": template_context,
        "template": template_name,
    }


django = {
    "co_names": ("get_response",),
    "path_fragment": "/kolo/middleware.py",
    "call_type": "django_request",
    "return_type": "django_response",
    "process": process_django,
}

django_async = {
    "co_names": ("aget_response",),
    "path_fragment": "/kolo/middleware.py",
    "call_type": "django_request",
    "return_type": "django_response",
    "process": process_django,
}

django_template = {
    "co_names": ("_render",),
    "path_fragment": "django/template/base.py",
    "call_type": "django_template_start",
    "return_type": "django_template_end",
    "process": process_django_template,
}

django_template_instrumented_test = {
    "co_names": ("instrumented_test_render",),
    "path_fragment": "django/test/utils.py",
    "call_type": "django_template_start",
    "return_type": "django_template_end",
    "process": process_django_template,
}

django_setup = {
    "co_names": ("setup",),
    "path_fragment": "django/__init__.py",
    "call_type": "django_setup_start",
    "return_type": "django_setup_end",
    "process": None,
}

django_checks = {
    "co_names": ("run_checks",),
    "path_fragment": "django/core/checks/registry.py",
    "call_type": "django_checks_start",
    "return_type": "django_checks_end",
    "process": None,
}

django_test_db = {
    "co_names": ("create_test_db",),
    "path_fragment": "django/db/backends/base/creation.py",
    "call_type": "django_create_test_db_start",
    "return_type": "django_create_test_db_end",
    "process": None,
}



from __future__ import annotations

import json
import logging
from datetime import datetime, timedelta
from functools import cached_property
from hashlib import sha1
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from typing import Optional
from urllib.parse import parse_qsl, urlparse

from ..config import load_config
from ..db import get_db_last_modified
from ..open_in_pycharm import OpenInPyCharmError
from ..settings import get_webapp_settings
from .api import (
    check_vacuum_needed,
    delete_trace,
    get_latest_version_info,
    get_trace,
    get_user_info,
    handle_open_editor,
    handle_static_file,
    handle_user_info_update,
    latest_traces,
    read_source_file,
    save_settings_data,
)
from .home import web_home_html

logger = logging.getLogger("kolo")


class KoloRequestHandler(BaseHTTPRequestHandler):
    protocol_version = "HTTP/1.1"

    def log_message(self, format, *args):
        # Skip logging for latest-traces endpoint
        if "/latest-traces/" in args[0]:
            return

        if "/static/" in args[0] or "favicon.ico" in args[0]:
            return

        if "/api/init/" in args[0] or "favicon.ico" in args[0]:
            return

        # Keep default logging for all other requests
        super().log_message(format, *args)

    @cached_property
    def url(self):
        return urlparse(self.path)

    @cached_property
    def query_data(self):
        return dict(parse_qsl(self.url.query))

    def set_cache_headers(
        self,
        max_age: int,
        etag: Optional[str] = None,
        last_modified: Optional[str | float | int | datetime] = None,
        immutable: bool = False,
    ) -> None:
        """Set common caching headers.

        Args:
            max_age: Cache duration in seconds
            etag: Optional ETag value
            last_modified: Optional timestamp (accepts datetime, string, float, or int)
            immutable: Whether to mark the response as immutable
        """
        cache_control = f"max-age={max_age}"
        if immutable:
            cache_control += ", immutable"
        self.send_header("Cache-Control", cache_control)
        if etag:
            self.send_header("ETag", etag)  # Django doesn't quote these either
        if last_modified is not None:
            if isinstance(last_modified, datetime):
                timestamp = last_modified.timestamp()
            elif isinstance(last_modified, (int, float)):
                timestamp = float(last_modified)
            else:
                # Try parsing as datetime string
                timestamp = datetime.fromisoformat(str(last_modified)).timestamp()
            self.send_header("Last-Modified", self.date_time_string(int(timestamp)))
            # Add Expires header to match Django
            expires = datetime.now() + timedelta(seconds=max_age)
            self.send_header("Expires", self.date_time_string(int(expires.timestamp())))

    def is_not_modified(self, etag: str) -> bool:
        """Check if the client's cached version is still valid."""
        if_none_match = self.headers.get("If-None-Match")
        return if_none_match == etag

    def send_not_modified(self) -> None:
        """Send a 304 Not Modified response."""
        self.send_response(HTTPStatus.NOT_MODIFIED)
        self.end_headers()

    def send_response_with_content(
        self,
        content: bytes,
        content_type: str,
        status: HTTPStatus = HTTPStatus.OK,
        max_age: Optional[int] = None,
        etag: Optional[str] = None,
        last_modified: Optional[str | float | int | datetime] = None,
        immutable: bool = False,
    ) -> None:
        """Send a response with proper HTTP headers."""
        self.send_response(status)
        self.send_header("Content-Type", content_type)
        self.send_header("Content-Length", str(len(content)))

        if max_age is not None:
            self.set_cache_headers(max_age, etag, last_modified, immutable)

        self.end_headers()
        self.wfile.write(content)
        self.wfile.flush()

    def send_error_response(self, code: HTTPStatus, message: str):
        """Send a complete error response with message."""
        # Use the parent class's send_response_only for the status line
        super().send_response_only(code, message)
        message_bytes = message.encode("utf-8")
        self.send_header("Content-Type", "text/plain")
        self.send_header("Content-Length", str(len(message_bytes)))
        self.send_header("Connection", "close")
        self.end_headers()
        self.wfile.write(message_bytes)

    def do_GET(self):
        if self.path.startswith("/_kolo/static/"):
            self.handle_static()
        elif self.path.startswith("/_kolo/api/traces/"):
            self.get_trace()
        elif self.path.startswith("/_kolo/api/latest-traces/"):
            self.latest_traces()
        elif self.path.startswith("/_kolo/api/source-file/"):
            self.get_source_file()
        elif self.path.startswith("/_kolo/api/config/"):
            self.get_config()
        elif self.path.startswith("/_kolo/api/init/"):
            self.init()
        elif self.path.startswith("/_kolo/api/settings/"):
            self.get_settings()
        elif self.path.startswith("/_kolo/api/open-editor/"):
            self.open_editor()
        elif self.path.startswith("/_kolo/api/user-info/"):
            self.get_user_info()
        elif self.path.startswith("/_kolo/api/latest-version/"):
            self.get_latest_version()
        elif self.path.startswith("/_kolo/api"):
            self.send_error_response(HTTPStatus.NOT_FOUND, "Kolo Web API: Not Found")
        elif self.path.startswith("/_kolo/"):
            self.web_home()
        else:
            self.send_response(HTTPStatus.PERMANENT_REDIRECT)
            self.send_header("Location", "/_kolo/")
            self.end_headers()

    def do_DELETE(self):
        if self.path.startswith("/_kolo/api/traces/"):
            self.delete_trace()
        else:
            self.send_error_response(HTTPStatus.NOT_FOUND, "Kolo Web API: Not Found")

    def do_POST(self):
        if self.path.startswith("/_kolo/api/settings/"):
            self.save_settings()
        elif self.path.startswith("/_kolo/api/user-info/"):
            self.save_user_info()
        else:
            self.send_error_response(HTTPStatus.NOT_FOUND, "Kolo Web API: Not Found")

    def web_home(self):
        html = web_home_html(running_in_user_django=False)

        self.send_response_with_content(
            content=html.encode("utf-8"), content_type="text/html; charset=utf-8"
        )

    def write_json(self, json_data):
        self.send_response_with_content(
            content=json_data.encode("utf-8"), content_type="application/json"
        )

    def get_trace(self):
        trace_id = self.path.replace("/_kolo/api/traces/", "").replace("/", "")
        msgpack_data, created_at = get_trace(trace_id)

        self.send_response_with_content(
            content=msgpack_data,
            content_type="application/msgpack",
            max_age=2419200,  # 28 days
            etag=trace_id,
            last_modified=created_at,
            immutable=True,  # Mark trace responses as immutable
        )

    def delete_trace(self):
        trace_id = self.path.replace("/_kolo/api/traces/", "").replace("/", "")
        deleted = delete_trace(trace_id)
        self.write_json(json.dumps(deleted))

    def latest_traces(self):
        last_modified = get_db_last_modified()
        etag = (
            sha1(last_modified.isoformat().encode("utf-8")).hexdigest()
            if last_modified
            else None
        )
        if etag and self.is_not_modified(etag):
            return self.send_not_modified()

        if "anchor" in self.query_data and "showNext" in self.query_data:
            anchor = self.query_data["anchor"]
            show_next = int(self.query_data["showNext"])
        else:
            anchor = None
            show_next = int(self.query_data.get("show_next", 10))

        trace_json = json.dumps(latest_traces(show_next, anchor))

        self.send_response_with_content(
            content=trace_json.encode("utf-8"),
            content_type="application/json",
            max_age=1,
            etag=etag,
            last_modified=last_modified.timestamp() if last_modified else None,
        )

    def get_source_file(self):
        path = self.query_data["path"]
        source = read_source_file(path)
        self.write_json(json.dumps(source))

    def handle_static(self):
        file_path = self.path[len("/_kolo/static/") :]
        try:
            contents, mime_type = handle_static_file(file_path)
            self.send_response_with_content(content=contents, content_type=mime_type)
        except FileNotFoundError:
            self.send_error_response(HTTPStatus.NOT_FOUND, "File not found")

    def get_config(self):
        config = load_config()
        self.write_json(json.dumps(config))

    def init(self):
        did_vacuum, last_vacuumed = check_vacuum_needed()
        self.write_json(
            json.dumps({"did_vacuum": did_vacuum, "last_vacuumed": last_vacuumed})
        )

    def get_settings(self):
        settings = get_webapp_settings()
        self.write_json(json.dumps(settings))

    def save_settings(self):
        content_length = int(self.headers["Content-Length"])
        body = self.rfile.read(content_length)
        settings = json.loads(body.decode("utf-8"))
        save_settings_data(settings)
        self.write_json(json.dumps({"ok": True}))

    def open_editor(self):
        file_path = self.query_data["path"]
        editor = self.query_data["editor"]
        try:
            handle_open_editor(file_path, editor)
            self.write_json(json.dumps({"ok": True}))
        except OpenInPyCharmError as e:
            self.send_error_response(HTTPStatus.INTERNAL_SERVER_ERROR, e.message)

    def get_user_info(self):
        self.write_json(json.dumps(get_user_info()))

    def save_user_info(self):
        content_length = int(self.headers["Content-Length"])
        body = self.rfile.read(content_length)
        data = json.loads(body.decode("utf-8"))
        email = data["email"]

        try:
            handle_user_info_update(email)
            self.write_json(json.dumps({"ok": True}))
        except ValueError as e:
            self.send_error_response(HTTPStatus.BAD_REQUEST, str(e))
        except IOError as e:
            logger.error("Could not save user details", e)
            self.send_error_response(
                HTTPStatus.INTERNAL_SERVER_ERROR, "Could not save user details"
            )

    def get_latest_version(self):
        try:
            version = get_latest_version_info()
            self.write_json(json.dumps({"latest_version": version}))
        except ValueError as e:
            self.send_error_response(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))



from __future__ import annotations

import logging
import types

formatter = logging.Formatter()


def process_logging(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    frame_locals = frame.f_locals
    logger_name = frame_locals["self"].name
    exc_info = frame_locals["exc_info"]
    traceback = None if exc_info is None else formatter.formatException(exc_info)
    extra = frame_locals["extra"]
    return {
        "args": frame_locals["args"],
        "extra": extra,
        "level": logging.getLevelName(frame_locals["level"]),
        "msg": frame_locals["msg"],
        "stack": formatter.formatStack(frame_locals["sinfo"]),
        "traceback": traceback,
        "name": logger_name,
    }


logging_plugin = {
    "co_names": ("_log",),
    "path_fragment": "/logging/",
    "events": ["return"],
    "call_type": "log_message",
    "return_type": "log_message",
    "process": process_logging,
}



from itertools import groupby
from operator import attrgetter

from tabulate import tabulate

from .queries import SkipQueryReason


def format_creates(model, creates):
    yield ""
    yield model
    queries = list(creates)
    first = queries[0]
    headers = ["variable name"]
    headers.extend([field.name for field in first.all_fields()])
    rows = []
    for query in queries:
        row = [query.variable_name]
        row.extend([field.value_repr for field in query.all_fields()])
        rows.append(row)
    yield tabulate(rows, headers=headers, disable_numparse=True)


def format_reason(reason):
    if reason is SkipQueryReason.NO_COLUMNS:
        yield "These queries did not select any columns:"
    elif reason is SkipQueryReason.NO_PRIMARY_KEY:
        yield "These queries did not select the primary key for a table:"
    elif reason is SkipQueryReason.NULL_PRIMARY_KEY:
        yield "These queries have a primary key set to NULL:"
    elif reason is SkipQueryReason.SEEN_MUTATION:
        yield "These queries select a row that has been mutated by an earlier query:"
    elif reason is SkipQueryReason.UNSUPPORTED_WHERE:
        yield "These queries use a WHERE clause Kolo doesn't yet handle:"


def format_skips(skips):
    for skip in skips:
        if skip.table:
            yield f"table: {skip.table}"
        yield f"query: {skip.query['query']}"
        yield f"query data: {skip.query['query_data']}"
        yield ""


def format_intermediate(context):
    for index, section in enumerate(context["sections"]):
        yield f"Section {index}"

        for model, creates in groupby(section["sql_fixtures"], key=attrgetter("model")):
            yield from format_creates(model, creates)

        yield "\nSkipped Queries:"

        skipped = sorted(section["skipped_inserts"], key=attrgetter("reason.value"))
        for reason, skips in groupby(skipped, key=attrgetter("reason")):
            yield from format_reason(reason)
            yield from format_skips(skips)



import types

import ulid


def build_context(config):
    return {
        "frame_ids": {},
        "request_ids": set(),
    }


def process_httpx(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    request_ids = context["request_ids"]
    frame_ids = context["frame_ids"]

    frame_locals = frame.f_locals
    # Track `request_id` to strip out unwanted resume frames in the async case.
    # Unfortunately `request_id` isn't guaranteed to be unique if the user chooses
    # to re-use a `request` instance, but I haven't thought of a way to do better.
    # We can't use `self` (`Client` or `AsyncClient` because that's even more likely
    # to be re-used.
    request_id = id(frame_locals["request"])
    request = frame_locals["request"]
    method = request.method
    url = str(request.url)

    if event in ("call", "resume"):
        if request_id in request_ids:
            frame_id = frame_ids[id(frame)]
        else:
            request_ids.add(request_id)
            frame_id = f"frm_{ulid.new()}"
            frame_ids[id(frame)] = frame_id
        return {
            "body": None,
            "frame_id": frame_id,
            "headers": dict(request.headers),
            "method": method,
            "method_and_full_url": f"{method} {url}",
            "url": url,
        }

    from httpx._models import Response

    if isinstance(arg, Response):
        request_ids.discard(request_id)

        from httpx._exceptions import ResponseNotRead

        try:
            body = arg.text
        except ResponseNotRead:  # pragma: no cover
            body = None
        headers = dict(arg.headers)
        status_code = arg.status_code
    else:
        # If we don't have a `Response`, we're suspending the coroutine
        # or there was an exception

        body = None
        headers = None
        status_code = None

    return {
        "body": body,
        "frame_id": frame_ids[id(frame)],
        "headers": headers,
        "method": method,
        "method_and_full_url": f"{method} {url}",
        "status_code": status_code,
        "url": url,
    }


httpx = {
    "co_names": ("send",),
    "path_fragment": "httpx/_client.py",
    "call_type": "outbound_http_request",
    "return_type": "outbound_http_response",
    "subtype": "httpx",
    "process": process_httpx,
    "build_context": build_context,
}



import json
from dataclasses import asdict
from importlib import import_module
from io import StringIO
from operator import attrgetter
from textwrap import indent
from typing import Any, Callable, ClassVar, Dict, List, Optional, Protocol, Tuple, Type

from . import factories, queries
from .steps import (
    AssertDelete,
    AssertInsert,
    AssertResponseJson,
    AssertStatusCode,
    AssertTemplateUsed,
    AssertUpdate,
    Code,
    CodeComment,
    DjangoTestClient,
    EmptyLine,
    EndTestClass,
    EndTestFunction,
    EndTestMethod,
    EndTimeTravel,
    FactoryCreate,
    Import,
    ModelCreate,
    ModelUpdate,
    ModuleImports,
    RegisterMocket,
    StartTimeTravel,
    Step,
    TestClass,
    TestFunction,
    TestMethod,
)


def json_encoder(obj):
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, Import):
        return {"import": obj.render()}
    raise TypeError(
        f"Object of type {obj.__class__.__name__} is not JSON serializable"
    )  # pragma: no cover


class Plan:
    interesting_steps = (
        AssertInsert,
        AssertResponseJson,
        AssertStatusCode,
        AssertTemplateUsed,
        AssertUpdate,
        DjangoTestClient,
        FactoryCreate,
        ModelCreate,
        ModelUpdate,
        RegisterMocket,
    )

    def __init__(self, steps, pytest, context):
        self.steps = list(steps)
        self.pytest = pytest
        self.context = context

    def render(self):
        contents = StringIO()

        imports = set()
        for step in self.steps:
            imports.update(step.get_imports(self.pytest))

        indent_count = 0
        for step in self.steps:
            indent_prefix = "    " * indent_count
            if isinstance(step, ModuleImports):
                for import_step in sorted(imports, key=attrgetter("import_path")):
                    contents.write(indent(import_step.render(), indent_prefix))

            contents.write(indent(step.render(self.pytest), indent_prefix))
            indent_count += step.indent_delta

        return contents.getvalue()

    def as_json(self):
        steps = [
            (step.__class__.__name__, asdict(step))
            for step in self.steps
            if isinstance(step, self.interesting_steps)
        ]
        return json.dumps(steps, default=json_encoder)


def header_comments(context):
    if context["now"]:
        yield CodeComment(f"Auto generated by Kolo at {context['now']}.")
    else:
        yield CodeComment("Auto generated by Kolo.")
    yield CodeComment(f"Kolo version: {context['kolo_version']}")
    if context["django_version"]:
        yield CodeComment(f"Django version: {context['django_version']}")
    for trace_id in context["_traces"]:
        yield CodeComment(f"Trace ID: {trace_id}")


def start_time_travel(section, context):
    if "request_timestamp" in section:
        request_timestamp = repr(section["request_timestamp"])
        if context["time_travel_tick"]:
            args: Tuple[str, ...] = (request_timestamp, "tick=True")
        else:
            args = (request_timestamp,)
        yield StartTimeTravel(
            context["time_travel_call"],
            args,
            imports={Import(context["time_travel_import"])},
        )


def end_time_travel(section):
    if "request_timestamp" in section:
        yield EndTimeTravel()


def call_test_client(section, pytest):
    if "response" in section:
        request = section["request"]
        post_data = request["post_data"]
        post_data.pop("csrfmiddlewaretoken", None)
        yield DjangoTestClient(
            request["method"].lower(),
            request["path_info"],
            request["query_params"],
            post_data,
            section["request_headers"],
        )
        yield EmptyLine()
        yield AssertStatusCode(section["response"]["status_code"])
        if section["response_json"]:
            yield AssertResponseJson(section["response_json"])
        for template_name in section["template_names"]:
            yield AssertTemplateUsed(template_name)
        yield EmptyLine()


def build_steps(context, pytest):
    yield from header_comments(context)
    yield EmptyLine()
    yield ModuleImports()
    yield EmptyLine()
    yield EmptyLine()
    if not pytest:
        yield TestClass(
            context["test_class"],
            ("TestCase",),
            {Import("from django.test import TestCase")},
        )
    yield Code("@mocketize(strict_mode=True)", {Import("from mocket import mocketize")})
    if pytest:
        yield Code("@pytest.mark.django_db()", {Import("import pytest")})
        yield TestFunction(context["test_name"], ("client",))
    else:
        yield TestMethod(context["test_name"])

    for section in context["sections"]:
        yield from start_time_travel(section, context)

        for create_fixture in section["sql_fixtures"]:
            if isinstance(create_fixture, queries.DjangoCreate):
                yield ModelCreate.from_fixture(create_fixture)
            elif isinstance(create_fixture, queries.DjangoUpdate):
                yield ModelUpdate.from_fixture(create_fixture)
            elif isinstance(create_fixture, factories.Factory):  # pragma: no branch
                yield FactoryCreate.from_fixture(create_fixture)
        if len(section["sql_fixtures"]) > 0:
            yield EmptyLine()

        for outbound_request in section["outbound_request_frames"]:
            yield RegisterMocket.from_outbound_request(outbound_request)
        if len(section["outbound_request_frames"]) > 0:
            yield EmptyLine()

        yield from call_test_client(section, pytest)

        for fixture in section["asserts"]:
            if isinstance(fixture, queries.AssertInsert):
                yield AssertInsert.from_fixture(fixture)
            elif isinstance(fixture, queries.AssertUpdate):
                yield AssertUpdate.from_fixture(fixture)
            elif isinstance(fixture, queries.AssertDelete):  # pragma: no branch
                yield AssertDelete.from_fixture(fixture)

        yield from end_time_travel(section)

    if pytest:
        yield EndTestFunction()
    else:
        yield EndTestMethod()
        yield EndTestClass()


def run_plan_hooks(plan, hooks):
    for hook in hooks:
        plan = hook(plan)
    return plan


def run_step_hooks(plan, hooks):
    for step_class, step_hooks in hooks.items():
        indexed_steps = step_class.get_steps(plan.steps)
        new_steps = []
        for index_slice, steps in indexed_steps:
            if len(steps) == 1:
                steps = steps[0]
            for hook in step_hooks:
                steps = hook(steps)
            new_steps.append((index_slice, steps))
        for index_slice, steps in reversed(new_steps):
            plan.steps[index_slice] = steps
    return plan


PlanHook = Callable[[Plan], Plan]
StepHook = Callable[[List[Step]], List[Step]]
plan_hooks: List[PlanHook] = []
step_hooks: Dict[Type[Step], List[StepHook]] = {}


def load_hooks(config):
    imports = config.get("test_generation", {}).get("hook_imports", [])
    for import_path in imports:
        import_module(import_path)  # pragma: no cover
    return plan_hooks, step_hooks


def plan_hook(func: PlanHook):
    plan_hooks.append(func)
    return func


def step_hook(step_class: Type[Step]):
    def decorator(func: StepHook):
        step_hooks.setdefault(step_class, []).append(func)
        return func

    return decorator



import sys
import threading


def get_third_party_profiler(config):
    profiler = sys.getprofile()
    if profiler:
        return profiler

    try:
        profiler = threading.getprofile()  # type: ignore[attr-defined]
    except AttributeError:
        profiler = threading._profile_hook
    if profiler:  # pragma: no branch
        return profiler

    return None



from importlib import import_module


def import_string(path):
    module_path, _sep, filter_name = path.rpartition(".")
    module = import_module(module_path)
    return getattr(module, filter_name)



from __future__ import annotations

import time
import types
from typing import Any


def call_sql(frame, event, arg, context):
    # Doing the import once and binding is substantially faster than going
    # through the import system every time.
    # See:
    # https://github.com/django/asgiref/issues/269
    # https://github.com/django/asgiref/pull/288
    # A more battle-tested (slightly slower/more complex) alternative would be:
    # https://github.com/django/django/pull/14850
    # https://github.com/django/django/pull/14858
    # https://github.com/django/django/pull/14931
    if "class" not in context:
        from django.db.models.sql.compiler import SQLUpdateCompiler

        context["class"] = SQLUpdateCompiler
    return frame.f_code is not context["class"].execute_sql.__code__


def process_sql(
    frame: types.FrameType,
    event: str,
    arg: Any,
    context,
):
    timestamp = time.time()
    database = frame.f_locals["self"].connection.vendor
    if event == "call":
        return {
            "database": database,
            "call_timestamp": timestamp,
        }

    assert event == "return"

    try:
        sql = frame.f_locals["sql"]
        params = frame.f_locals["params"]
    except KeyError:
        query_template = None
        query = None
        params = ()
    else:
        if sql == "":
            query_template = None
            query = None
        else:
            cursor = frame.f_locals["cursor"]
            ops = frame.f_locals["self"].connection.ops
            query_template = sql.strip()
            query = ops.last_executed_query(cursor, sql, params).strip()

    return {
        "database": database,
        "return_timestamp": timestamp,
        "query_template": query_template,
        "query": query,
        "query_data": arg,
        "query_params": params,
    }


def process_raw_sql(
    frame: types.FrameType,
    event: str,
    arg: Any,
    context,
):  # pragma: no cover
    timestamp = time.time()
    raw_query = frame.f_locals["self"]
    connection = frame.f_globals["connections"][raw_query.using]
    database = connection.vendor
    if event == "call":
        return {
            "database": database,
            "call_timestamp": timestamp,
        }

    assert event == "return"

    try:
        sql = raw_query.sql
        params = frame.f_locals["params"]
    except KeyError:
        query_template = None
        query = None
        params = ()
    else:
        cursor = raw_query.cursor
        ops = connection.ops
        query_template = sql.strip()
        query = ops.last_executed_query(cursor, sql, params).strip()

    return {
        "database": database,
        "return_timestamp": timestamp,
        "query_template": query_template,
        "query": query,
        "query_data": arg,
        "query_params": params,
    }


sql = {
    "co_names": ("execute_sql",),
    "path_fragment": "/django/db/models/sql/compiler.py",
    "call": call_sql,
    "call_type": "start_sql_query",
    "return_type": "end_sql_query",
    "process": process_sql,
}


django_raw_sql = {
    "co_names": ("_execute_query",),
    "path_fragment": "/django/db/models/sql/query.py",
    "call_type": "start_sql_query",
    "return_type": "end_sql_query",
    "process": process_raw_sql,
}



from __future__ import annotations

import ulid


def call_unittest(frame, event, arg, context):
    return (  # pragma: no cover
        event == "call" and frame.f_code.co_name == "startTest"
    ) or (event == "return" and frame.f_code.co_name == "stopTest")


def process_unittest(frame, event, arg, context):  # pragma: no cover
    frame_ids = context["frame_ids"]
    testcase = frame.f_locals["test"]
    co_name = frame.f_code.co_name
    if co_name == "startTest":
        frame_id = f"frm_{ulid.new()}"
        frame_ids[id(testcase)] = frame_id
    else:
        frame_id = frame_ids[id(testcase)]
    return {
        "frame_id": frame_id,
        "test_name": testcase._testMethodName,
        "test_class": testcase.__class__.__qualname__,
    }


def build_context(config):
    return {"frame_ids": {}}


unittest = {
    "co_names": ("startTest", "stopTest"),
    "path_fragment": "unittest/result.py",
    "call": call_unittest,
    "call_type": "start_test",
    "return_type": "end_test",
    "process": process_unittest,
    "build_context": build_context,
}



from __future__ import annotations

import json
import logging
import sqlite3
from contextlib import contextmanager
from datetime import date, datetime, time, timezone
from pathlib import Path
from typing import Any, Dict, Iterator, Tuple

from .config import create_kolo_directory
from .serialize import dump_msgpack, load_msgpack

logger = logging.getLogger("kolo")


class SchemaNotFoundError(Exception):
    pass


class TraceNotFoundError(Exception):
    pass


@contextmanager
def db_connection(db_path, timeout=60):
    """
    Wrap sqlite's connection for use as a context manager

    Commits all changes if no exception is raised.
    Always closes the connection after the context manager exits.
    """
    connection = sqlite3.connect(str(db_path), isolation_level=None, timeout=timeout)
    try:
        connection.execute("pragma journal_mode=wal")
    finally:
        connection.close()

    connection = sqlite3.connect(str(db_path), isolation_level=None, timeout=timeout)
    try:
        with connection:
            yield connection
    finally:
        connection.close()


def get_db_path() -> Path:
    return create_kolo_directory() / "db.sqlite3"


def get_db_last_modified() -> datetime | None:
    try:
        modified = get_db_path().stat().st_mtime_ns
    except FileNotFoundError:
        return None
    else:
        return datetime.fromtimestamp(modified / 1e9, tz=timezone.utc)


def create_traces_table(connection) -> None:
    create_table_query = """
    CREATE TABLE IF NOT EXISTS traces (
        id text PRIMARY KEY NOT NULL,
        created_at TEXT DEFAULT (STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')) NOT NULL,
        data text NULL,
        msgpack blob NULL,
        is_pinned INTEGER DEFAULT 0
    );
    """
    create_timestamp_index_query = """
        CREATE INDEX IF NOT EXISTS
        idx_traces_created_at
        ON traces (created_at);
        """

    connection.execute(create_table_query)
    connection.execute(create_timestamp_index_query)


def setup_db() -> Path:
    db_path = get_db_path()

    with db_connection(db_path) as connection:
        create_traces_table(connection)

    return db_path


def save_trace_in_sqlite(
    db_path: Path,
    trace_id: str,
    msgpack: bytes,
    *,
    ignore_errors: bool = True,
    created_at: datetime | None = None,
    timeout=60,
) -> None:
    ignore = " OR IGNORE" if ignore_errors else ""
    _columns = ["id", "msgpack"]
    values: list[object] = [trace_id, msgpack]
    if created_at is not None:
        _columns.append("created_at")
        values.append(created_at)
    columns = ", ".join(_columns)
    params = ",".join(["?" for _ in _columns])

    insert_sql = f"INSERT{ignore} INTO traces({columns}) VALUES({params})"

    # We can't reuse a connection
    # because we're in a new thread
    with db_connection(db_path, timeout) as connection:
        try:
            connection.execute(insert_sql, values)
        except (sqlite3.DataError, sqlite3.InterfaceError):
            # DataError on python 3.11+
            # InterfaceError on python 3.10-
            logger.exception("The generated trace was too big to store in sqlite.")


def load_trace_from_db(db_path: Path, trace_id: str) -> Tuple[bytes, str]:
    fetch_sql = "SELECT msgpack, created_at FROM traces WHERE id = ?"

    with db_connection(db_path) as connection:
        cursor = connection.execute(fetch_sql, (trace_id,))
        row = cursor.fetchone()
    if row is None:
        raise TraceNotFoundError(trace_id)
    return row


def load_trace_with_size_from_db(
    db_path: Path, trace_id: str
) -> Tuple[str, str, int, bytes]:
    fetch_sql = """
        SELECT id, created_at, LENGTH(msgpack), msgpack
        FROM traces WHERE id = ?
    """

    with db_connection(db_path) as connection:
        cursor = connection.execute(fetch_sql, (trace_id,))
        row = cursor.fetchone()
    if row is None:
        raise TraceNotFoundError(trace_id)
    return row


def list_traces_from_db(db_path: Path, count=500, reverse=False):
    list_sql = """
    SELECT id, created_at, LENGTH(msgpack)
    FROM traces ORDER BY id DESC LIMIT ?
    """

    with db_connection(db_path) as connection:
        cursor = connection.execute(list_sql, [count])
        rows = cursor.fetchall()
    if reverse:
        return reversed(rows)
    return rows


def list_traces_with_data_from_db(db_path: Path, count=500, reverse=False):
    """Like list_traces_from_db but includes the msgpack data for each trace."""
    list_sql = """
    SELECT id, created_at, LENGTH(msgpack), msgpack
    FROM traces ORDER BY id DESC LIMIT ?
    """

    with db_connection(db_path) as connection:
        cursor = connection.execute(list_sql, [count])
        rows = []
        while True:
            row = cursor.fetchone()
            if row is None:
                break
            if reverse:
                rows.append(row)
            else:
                yield row

    if reverse:
        for row in reversed(rows):
            yield row


def get_pinned_traces(db_path: Path) -> Iterator[Tuple[str, str, int, bytes]]:
    """Get all pinned traces from the database."""
    with db_connection(db_path) as connection:
        cursor = connection.execute(
            """
            SELECT id, created_at, LENGTH(msgpack), msgpack
            FROM traces
            WHERE is_pinned = 1
            ORDER BY id DESC
            """
        )
        while True:
            row = cursor.fetchone()
            if row is None:
                break
            yield row


def delete_traces_by_id(db_path: Path, trace_ids: Tuple[str, ...]):
    params = ", ".join("?" * len(trace_ids))
    delete_sql = f"DELETE FROM traces WHERE id in ({params})"

    with db_connection(db_path) as connection:
        cursor = connection.execute(delete_sql, trace_ids)
        return cursor.rowcount


def delete_traces_before(db_path: Path, before: datetime):
    delete_sql = "DELETE FROM traces WHERE (created_at < ?)"

    with db_connection(db_path) as connection:
        connection.execute(delete_sql, (before,))
        cursor = connection.execute("SELECT changes()")
        deleted_count = cursor.fetchone()[0]
    return deleted_count


def vacuum_db(db_path):
    with db_connection(db_path) as connection:
        connection.execute("VACUUM")


def create_schema_table(connection) -> None:
    create_table_query = """
    CREATE TABLE IF NOT EXISTS schemas (
        id integer PRIMARY KEY,
        created_at TEXT DEFAULT (STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')) NOT NULL,
        git_commit TEXT NULL,
        data text NOT NULL
    );
    """

    connection.execute(create_table_query)


class SchemaEncoder(json.JSONEncoder):
    def default(self, obj):
        from django.utils.functional import Promise

        if isinstance(obj, Promise):
            return str(obj)
        if isinstance(obj, datetime):
            return {"__type": "datetime", "value": obj.isoformat()}
        if isinstance(obj, date):
            return {"__type": "date", "value": obj.isoformat()}
        if isinstance(obj, time):
            return {"__type": "time", "value": obj.isoformat()}
        return super().default(obj)  # pragma: no cover


def decode_value(obj):
    if "__type" not in obj:
        return obj
    _type = obj["__type"]
    value = obj["value"]
    if _type == "datetime":
        return datetime.fromisoformat(value)
    if _type == "date":
        return date.fromisoformat(value)
    if _type == "time":
        return time.fromisoformat(value)
    raise ValueError(f"Could not decode: {obj}")  # pragma: no cover


def save_schema(connection, schema, commit_sha) -> None:
    delete_sql = "DELETE FROM schemas WHERE git_commit = ?"
    insert_sql = 'INSERT INTO schemas("data", "git_commit") VALUES(?, ?)'

    json_schema = json.dumps(schema, cls=SchemaEncoder)
    connection.execute(delete_sql, (commit_sha,))
    connection.execute(insert_sql, (json_schema, commit_sha))


def load_schema(db_path: Path) -> Tuple[Dict[str, Any], str]:
    load_sql = """SELECT data, git_commit FROM schemas
        ORDER BY created_at DESC LIMIT 1"""

    with db_connection(db_path) as connection:
        cursor = connection.execute(load_sql)
        data, commit = cursor.fetchone()

    data = json.loads(data, object_hook=decode_value)
    return data, commit


def load_schema_for_commit_sha(
    db_path: Path, commit_sha: str
) -> Tuple[Dict[str, Any], str]:
    load_sql = """SELECT data FROM schemas WHERE git_commit = ?
        ORDER BY created_at DESC LIMIT 1"""

    with db_connection(db_path) as connection:
        try:
            cursor = connection.execute(load_sql, [commit_sha])
        except sqlite3.OperationalError:
            row = None
        else:
            row = cursor.fetchone()

    if not row:
        raise SchemaNotFoundError(commit_sha)

    return json.loads(row[0], object_hook=decode_value)


def convert_json_to_msgpack(db_path: Path):  # pragma: no cover
    json_traces = "SELECT id, data FROM traces WHERE data IS NOT NULL"
    update_trace = "UPDATE traces SET data = NULL, msgpack = ? WHERE id = ?"

    with db_connection(db_path) as connection:
        cursor = connection.execute(json_traces)
        rows = cursor.fetchall()

        for trace_id, json_data in rows:
            msgpack_data = dump_msgpack(json.loads(json_data))
            cursor.execute(update_trace, (msgpack_data, trace_id))

    return len(rows)


def pin_trace(db_path: Path, trace_id: str) -> bool:
    """Pin a trace. Returns True if the trace was found and pinned."""
    update_sql = """
    UPDATE traces 
    SET is_pinned = 1
    WHERE id = ?
    """

    with db_connection(db_path) as connection:
        try:
            connection.execute(
                "ALTER TABLE traces ADD COLUMN is_pinned INTEGER DEFAULT 0"
            )
        except sqlite3.OperationalError:
            # Column already exists
            pass

        cursor = connection.execute(update_sql, (trace_id,))
        return cursor.rowcount > 0


def unpin_trace(db_path: Path, trace_id: str) -> bool:
    """Unpin a trace. Returns True if the trace was found and unpinned."""
    update_sql = """
    UPDATE traces 
    SET is_pinned = 0
    WHERE id = ?
    """

    with db_connection(db_path) as connection:
        cursor = connection.execute(update_sql, (trace_id,))
        return cursor.rowcount > 0



import subprocess
import sys
from tempfile import NamedTemporaryFile

import msgpack


def pretty_byte_size(size_bytes):
    if size_bytes < 1024:
        return f"{size_bytes:5d} B"

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    bit_length = size_bytes.bit_length() - 1
    index = bit_length // 10
    denominator = 1 << (10 * index)
    ratio = size_bytes / denominator
    return f"{ratio:5.1f} {size_name[index]}"


def maybe_format(rendered):
    rendered = maybe_isort(rendered)

    try:
        import ruff
    except ImportError:
        if len(rendered) <= 1_000_000:
            rendered = maybe_black(rendered)
        return rendered

    with NamedTemporaryFile("w+", delete=False) as f:
        f.write(rendered)
        name = f.name

    subprocess.run(
        [
            sys.executable,
            "-m",
            "ruff",
            "format",
            name,
            "--config",
            "format.skip-magic-trailing-comma=true",
        ],
        check=True,
    )
    with open(name) as f:
        rendered = f.read()

    return rendered


def maybe_black(rendered):
    try:
        from black import format_file_contents
        from black.mode import Mode
        from black.parsing import InvalidInput
        from black.report import NothingChanged
    except ImportError:
        return rendered

    try:
        return format_file_contents(
            rendered, fast=True, mode=Mode(magic_trailing_comma=False)
        )
    except (InvalidInput, NothingChanged):
        return rendered


def maybe_isort(rendered):
    try:
        from isort.api import sort_code_string
    except ImportError:
        return rendered
    return sort_code_string(rendered)


def get_terminal_formatter(mode):
    if mode == "off":
        return None

    import os

    from pygments.formatters import (
        Terminal256Formatter,
        TerminalFormatter,
        TerminalTrueColorFormatter,
    )

    style = "monokai" if mode == "dark" else "default"

    # Derived from https://github.com/pygments/pygments/blob/e49aef678a5ce40b6ed7b38cd71dddafc12f979c/pygments/cmdline.py#L448-L453
    if os.environ.get("COLORTERM", "") in ("truecolor", "24bit"):
        return TerminalTrueColorFormatter(style=style)
    if "256" in os.environ.get("TERM", ""):
        return Terminal256Formatter(style=style)
    return TerminalFormatter(bg=mode)


def get_sql_lexer(dialect):
    from pygments.lexers import MySqlLexer, PostgresLexer, SqlLexer

    if dialect == "postgres":
        return PostgresLexer()
    if dialect == "mysql":
        return MySqlLexer()
    return SqlLexer()


def highlight_sql(query, dialect, formatter):
    if formatter is None:
        # pygments adds an extra newline while highlighting.
        # Since this looks better, we also add one ourselves here.
        return query + "\n"

    from pygments import highlight

    return highlight(query, get_sql_lexer(dialect), formatter)


def highlight_python(code, formatter):
    from pygments import highlight
    from pygments.lexers import PythonLexer

    return highlight(code, PythonLexer(), formatter)


def extract_main_frames_from_data(data):
    if "config" not in data["meta"]:
        # config is not present in old traces
        return data["frames_of_interest"]

    if data.get("threads"):
        current_thread_id = data["current_thread_id"]
        if current_thread_id in data["threads"]:
            return data["threads"][current_thread_id]["frames"]
        else:
            return []
    else:
        return data["frames_of_interest"]


def extract_http_trace_name(frames_by_thread, current_thread_id):
    """
    Extract HTTP request/response information from frames to set a trace name.
    Looks for django_request and django_response frame types from the Django filter.

    Returns:
        A formatted trace name string or None if no HTTP information found
    """
    request_frame = None
    response_frame = None

    frames_from_current_thread = frames_by_thread.get(current_thread_id, [])
    first_three = frames_from_current_thread[:3]
    last_three = frames_from_current_thread[-3:]

    relevant_frames = first_three + last_three

    unpacked_frames = [
        msgpack.unpackb(f, strict_map_key=False) for f in relevant_frames
    ]

    for frame in unpacked_frames:
        if frame.get("type") == "django_request":
            print("extracting http trace name - django request")
            if request_frame is None:
                # first frame wins
                request_frame = frame
        elif frame.get("type") == "django_response":
            # last frame wins
            response_frame = frame

    if request_frame and response_frame:
        method = request_frame.get("method")
        path = request_frame.get("path_info")
        status_code = response_frame.get("status_code")

        if method and path and status_code:
            return f"{status_code} {method} {path}"

    return None



from __future__ import annotations

import types
from typing import TYPE_CHECKING

from ..serialize import decode_body


def process_requests(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    request = frame.f_locals["request"]
    method_and_url = f"{request.method} {request.url}"

    if event == "call":
        return {
            "body": decode_body(request.body, request.headers),
            "headers": dict(request.headers),
            "method": request.method,
            "method_and_full_url": method_and_url,
            "url": request.url,
        }

    assert event == "return"

    response = arg
    if TYPE_CHECKING:
        from requests.models import Response

        assert isinstance(response, Response)

    if response is None:
        body = None
        headers = None
        status_code = None
    else:
        body = response.text
        headers = dict(response.headers)
        status_code = response.status_code

    return {
        "body": body,
        "headers": headers,
        "method": request.method,
        "method_and_full_url": method_and_url,
        "status_code": status_code,
        "url": request.url,
    }


requests_plugin = {
    "co_names": ("send",),
    "path_fragment": "requests/sessions",
    "call_type": "outbound_http_request",
    "return_type": "outbound_http_response",
    "subtype": "requests",
    "process": process_requests,
}



import threading


def print_thread_debug_info():
    main_thread = threading.main_thread()
    current_thread = threading.current_thread()

    print(
        f"main_thread: {main_thread.name} ({main_thread.native_id})",
    )
    print(
        f"current_thread: {current_thread.name} ({current_thread.native_id})",
    )
    print("other threads:")
    for thread in threading.enumerate():
        if thread is not main_thread and thread is not current_thread:
            print(f"  {thread.name} ({thread.native_id})")


def get_thread_id(thread):
    native_id = getattr(thread, "native_id", None)
    ident = getattr(thread, "ident", None)

    if native_id:
        return f"native_{native_id}"
    elif ident:
        return f"ident_{ident}"
    else:
        print(f"Kolo warning: thread has no id: {getattr(thread, 'name', 'unknown')}")
        return "no_thread_id"



from __future__ import annotations


def process_huey(frame, event, arg, context):
    frame_locals = frame.f_locals
    task_object = frame_locals["self"]
    task_args, task_kwargs = task_object.data

    return {
        "name": f"{task_object.__module__}.{task_object.name}",
        "args": task_args,
        "kwargs": task_kwargs,
    }


def call_huey(frame, event, arg, context):
    # Doing the import once and binding is substantially faster than going
    # through the import system every time.
    # See:
    # https://github.com/django/asgiref/issues/269
    # https://github.com/django/asgiref/pull/288
    # A more battle-tested (slightly slower/more complex) alternative would be:
    # https://github.com/django/django/pull/14850
    # https://github.com/django/django/pull/14858
    # https://github.com/django/django/pull/14931
    if "class" not in context:
        from huey.api import Task

        context["class"] = Task
    return isinstance(frame.f_locals["self"], context["class"])


huey = {
    "co_names": ("execute",),
    "path_fragment": "/huey/api.py",
    "call": call_huey,
    "call_type": "background_job",
    "return_type": "background_job_end",
    "subtype": "huey",
    "process": process_huey,
}



import gzip
import logging
import os
from typing import Optional

import httpx

logger = logging.getLogger("kolo")


def upload_to_dashboard(trace: bytes, upload_key: Optional[str] = None):
    # This should be authed with the upload token optionally..

    base_url = os.environ.get("KOLO_BASE_URL", "https://my.kolo.app")
    url = f"{base_url}/api/traces/"
    payload = gzip.compress(trace)

    if upload_key:
        headers = {"Authorization": f"Bearer {upload_key}"}
    else:
        headers = {}
    return httpx.post(url, headers=headers, files={"data": payload})



from __future__ import annotations

import contextlib
import inspect
import io
import json
import logging
import os
import sys
from typing import Awaitable, Callable, List

from asgiref.sync import iscoroutinefunction, markcoroutinefunction, sync_to_async
from django.conf import settings
from django.core.servers.basehttp import ServerHandler
from django.http import (
    FileResponse,
    HttpRequest,
    HttpResponse,
    HttpResponseNotFound,
    JsonResponse,
)

from .checks import get_third_party_profiler
from .config import load_config
from .db import setup_db
from .open_in_pycharm import OpenInPyCharmError
from .profiler import enable
from .serialize import monkeypatch_queryset_repr
from .settings import get_webapp_settings
from .web.api import (
    check_vacuum_needed,
    get_latest_version_info,
    get_project_root,
    get_user_info,
    handle_open_editor,
    handle_static_file,
    handle_user_info_update,
    save_settings_data,
)
from .web.django import (
    kolo_web_api_delete_trace,
    kolo_web_api_generate_test,
    kolo_web_api_get_trace,
    kolo_web_api_latest_traces,
    kolo_web_api_source_file,
    kolo_web_home,
)

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]
DjangoAsyncView = Callable[[HttpRequest], Awaitable[HttpResponse]]


def get_host(args):
    if len(args) < 2:
        return ""

    if args[1] != "runserver":
        return ""

    ipv6 = "-6" in args or "--ipv6" in args
    if ipv6:
        host = "[::1]"
    else:
        host = "127.0.0.1"

    args = [arg for arg in args[2:] if not arg.startswith("-")]

    if not args:
        port = 8000
    else:
        _host, sep, port = args[0].rpartition(":")
        if sep:
            host = _host

    return f"http://{host}:{port}"


class KoloMiddleware:
    sync_capable = True
    async_capable = True

    def __init__(self, get_response: DjangoView | DjangoAsyncView) -> None:
        self._is_coroutine = iscoroutinefunction(get_response)
        if self._is_coroutine:
            markcoroutinefunction(self)
        self._get_response = get_response
        self.config = load_config()
        if settings.DEBUG:
            self.upload_token = None
        else:
            self.upload_token = self.get_upload_token()
        self.enabled = self.should_enable()
        if self.enabled:
            self.db_path = setup_db()

            # TODO: Put the full URL here not just the /_kolo/ path
            if not self.config.get("hide_startup_message", False):
                host = get_host(sys.argv)
                print(f"\nView recent requests at {host}/_kolo/")

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if request.path.startswith("/_kolo"):
            if self._is_coroutine:
                hide_from_daphne()
                return sync_to_async(kolo_web_router)(request)  # type: ignore[return-value]
            return kolo_web_router(request)

        if not self._is_coroutine:
            get_response = self.get_response
        else:
            get_response = self.aget_response  # type: ignore

        # WARNING: Because Django's runserver uses threading, we need
        # to be careful about thread safety here.
        if not self.enabled or self.check_for_third_party_profiler():
            return get_response(request)

        filter_config = self.config.get("filters", {})
        ignore_request_paths = filter_config.get("ignore_request_paths", [])
        for path in ignore_request_paths:
            if path in request.path:
                return get_response(request)

        monkeypatch_queryset_repr()
        if self._is_coroutine:
            return self.aprofile_response(request)
        else:
            return self.profile_response(request)

    def profile_response(self, request):
        with enable(
            self.config,
            source="kolo.middleware.KoloMiddleware",
            _save_in_thread=True,
            _upload_token=self.upload_token,
        ):
            response = self.get_response(request)
        return response

    async def aprofile_response(self, request):
        with enable(
            self.config,
            source="kolo.middleware.KoloMiddleware",
            _save_in_thread=True,
            _upload_token=self.upload_token,
        ):
            response = await self.aget_response(request)
        return response

    async def aget_response(self, request: HttpRequest) -> HttpResponse:
        response = await self._get_response(request)  # type: ignore
        return response

    def get_response(self, request: HttpRequest) -> HttpResponse:
        response = self._get_response(request)
        return response  # type: ignore

    def check_for_third_party_profiler(self) -> bool:
        profiler = get_third_party_profiler(self.config)
        if profiler:
            logger.warning("Profiler %s is active, disabling KoloMiddleware", profiler)
            return True
        return False

    def should_enable(self) -> bool:
        if settings.DEBUG is False and self.upload_token is None:
            logger.debug("DEBUG mode is off, disabling KoloMiddleware")
            return False

        if os.environ.get("KOLO_DISABLE", "false").lower() not in ["false", "0"]:
            logger.debug("KOLO_DISABLE is set, disabling KoloMiddleware")
            return False

        return not self.check_for_third_party_profiler()

    def get_upload_token(self):
        if not self.config.get("production_beta", False):
            return None

        upload_token = os.environ.get("KOLO_API_TOKEN", None)
        if upload_token is None:
            logging.warning(
                "Kolo production beta is enabled, but `KOLO_API_TOKEN` environment variable is not set."
            )
            return None

        if upload_token.startswith("kolo_prod_"):
            return upload_token

        logging.warning("`KOLO_API_TOKEN` is invalid.")
        return None


@contextlib.contextmanager
def hide_from_runserver(*args, **kwds):
    """
    Hides the requestline log messages from runserver's stdout.
    This works because Django's runserver is built on `wsgiref` which is ultimately built on `TCPServer` and the notion
    of a "server" creating a class "per request" which means we can rely on there being one `WSGIRequestHandler`
    for every incoming request, and we can modify that instance's methods to silence it.

    We don't want to restore the original method on exiting the context manager, because that would be "too soon" and
    the log message would ultimately still get spooled out at the WSGI server layer later (i.e. higher up the callstack)
    """

    def no_log(*a, **kw):
        return None

    for frame in inspect.stack():
        if "self" in frame.frame.f_locals:
            if isinstance(frame.frame.f_locals["self"], ServerHandler):
                server_handler = frame.frame.f_locals["self"]
                if hasattr(server_handler, "request_handler"):
                    server_handler.request_handler.log_message = no_log
    yield


DAPHNE_PATCHED = False


def hide_from_daphne(*args, **kwds):
    """
    Hides the requestline log messages from daphne runserver's stdout.
    """

    global DAPHNE_PATCHED

    if DAPHNE_PATCHED:
        return

    from daphne.server import Server

    for frame in inspect.stack():
        if "self" in frame.frame.f_locals:
            server = frame.frame.f_locals["self"]
            if isinstance(server, Server):
                old_log = server.action_logger

                def no_log(protocol, action, details):
                    if details["path"].startswith("/_kolo"):
                        return None
                    return old_log(protocol, action, details)

                server.action_logger = no_log
                DAPHNE_PATCHED = True
                break


@hide_from_runserver()
def kolo_web_router(request: HttpRequest) -> HttpResponse:
    request.get_host()  # Trigger any ALLOWED_HOSTS error
    path = request.path

    # Static paths
    if path.startswith("/_kolo/static/"):
        try:
            contents, mime_type = handle_static_file(path[len("/_kolo/static/") :])
            return FileResponse(
                io.BytesIO(contents),
                content_type=mime_type,
            )  # type: ignore # FileResponse inherits from StreamingHttpResponse but hard to fix the downstream errors when fixing that type
        except FileNotFoundError:
            return HttpResponseNotFound("File not found")

    # API paths
    elif path.startswith("/_kolo/api"):
        if path.startswith("/_kolo/api/generate-test/"):
            return kolo_web_api_generate_test(request)
        elif path.startswith("/_kolo/api/traces/"):
            if request.method == "GET":
                return kolo_web_api_get_trace(request)
            elif request.method == "DELETE":
                return kolo_web_api_delete_trace(request)
            else:
                return HttpResponseNotFound("Kolo Web: Not Found")
        elif path.startswith("/_kolo/api/latest-traces/"):
            return kolo_web_api_latest_traces(request)
        elif path.startswith("/_kolo/api/save-test/"):
            return kolo_web_api_save_test(request)
        elif path.startswith("/_kolo/api/config/"):
            return kolo_web_api_config(request)
        elif path.startswith("/_kolo/api/init/"):
            return kolo_web_api_init(request)
        elif path.startswith("/_kolo/api/settings/"):
            return kolo_web_api_settings(request)
        elif path.startswith("/_kolo/api/source-file/"):
            return kolo_web_api_source_file(request)
        elif path.startswith("/_kolo/api/open-editor/"):
            return kolo_web_api_open_editor(request)
        elif path.startswith("/_kolo/api/user-info/"):
            return kolo_web_api_user_info(request)
        elif path.startswith("/_kolo/api/latest-version/"):
            return kolo_web_api_latest_version(request)
        else:
            return HttpResponseNotFound("Kolo Web API: Not Found")

    # SPA path (let React render and handle the path)
    else:
        return kolo_web_home(request)


def kolo_web_api_save_test(request: HttpRequest) -> HttpResponse:
    project_folder = get_project_root()
    data = json.loads(request.body.decode("utf-8"))
    file_content: str = data["content"]
    relative_file_path: List[str] = data["path"]
    file_path = project_folder / os.path.join(*relative_file_path)

    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, "w") as file:
        file.write(file_content)

    return JsonResponse({"file_path": f"{file_path}"})


def kolo_web_api_config(request: HttpRequest) -> HttpResponse:
    return JsonResponse(load_config())


def kolo_web_api_init(request: HttpRequest) -> HttpResponse:
    did_vacuum, last_vacuumed = check_vacuum_needed()
    return JsonResponse({"did_vacuum": did_vacuum, "last_vacuumed": last_vacuumed})


def kolo_web_api_settings(request: HttpRequest) -> HttpResponse:
    """
    Return the `webapp_settings` stored in the `kolo_kv` SQLite table.
    """
    if request.method == "GET":
        return JsonResponse(get_webapp_settings())
    elif request.method == "POST":
        settings = json.loads(request.body.decode("utf-8"))
        save_settings_data(settings)
        return JsonResponse({"ok": True})
    else:
        return JsonResponse({"ok": False}, status=404)


def kolo_web_api_open_editor(request: HttpRequest) -> HttpResponse:
    file_path = request.GET["path"]
    editor = request.GET["editor"]
    try:
        handle_open_editor(file_path, editor)
        return JsonResponse({"ok": True})
    except OpenInPyCharmError as e:
        return JsonResponse({"ok": False, "message": e.message}, status=500)


def kolo_web_api_user_info(request: HttpRequest) -> JsonResponse:
    if request.method == "GET":
        return JsonResponse(get_user_info())
    elif request.method == "POST":
        data = json.loads(request.body.decode("utf-8"))
        email = data["email"]
        try:
            handle_user_info_update(email)
            return JsonResponse({"ok": True})
        except ValueError as e:
            return JsonResponse({"ok": False, "message": str(e)}, status=400)
        except IOError as e:
            logger.error("Could not save user details", e)
            return JsonResponse(
                {"ok": False, "message": "Could not save user details"}, status=500
            )
    else:
        return JsonResponse({"ok": False}, status=404)


def kolo_web_api_latest_version(request: HttpRequest) -> JsonResponse:
    try:
        version = get_latest_version_info()
        return JsonResponse({"latest_version": version})
    except ValueError as e:
        return JsonResponse({"ok": False, "message": str(e)}, status=500)



import importlib
import os
import sys
import time
from typing import Optional

import ulid

from .serialize import user_code_call_site

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points


def load(processor):
    mod, sep, name = processor.partition(":")
    processor = importlib.import_module(mod)
    if sep:  # pragma: no branch
        for attr in name.split("."):
            processor = getattr(processor, attr)

    return processor


def load_plugin_data(config):
    kolo_plugins = entry_points(group="kolo.processors")
    processors = [plugin.load() for plugin in kolo_plugins]
    local_procesors = config.get("processors", [])
    processors.extend([load(processor) for processor in local_procesors])
    return processors


class PluginProcessor:
    def __init__(self, plugin_data, config):
        self.plugin_data = plugin_data
        self.config = config

        self.co_names = plugin_data["co_names"]
        self.filename = os.path.normpath(plugin_data["path_fragment"])
        self.call_type = plugin_data["call_type"]
        self.return_type = plugin_data["return_type"]
        self.unwind_type = plugin_data.get("unwind_type", self.return_type)
        self.resume_type = plugin_data.get("resume_type", self.call_type)
        self.yield_type = plugin_data.get("yield_type", self.return_type)
        self.throw_type = plugin_data.get("throw_type", self.unwind_type)
        self.subtype = plugin_data.get("subtype")
        self.call_extra = plugin_data.get("call")
        self.process_extra = plugin_data.get("process")
        self.events = plugin_data.get("events")
        build_context = plugin_data.get("build_context")
        if build_context is None:
            self.context = {}
        else:
            self.context = build_context(config)

        self.frame_ids = {}

    def __call__(self, frame, event, arg):
        filename = frame.f_code.co_filename
        co_name = frame.f_code.co_name

        if self.call_extra is None:
            return co_name in self.co_names and self.filename in filename
        return (
            co_name in self.co_names
            and self.filename in filename
            and self.call_extra(frame, event, arg, self.context)
        )

    def matches(self, filename, name):
        return name in self.co_names and self.filename in filename

    def frame_id(self, frame, event):
        frame_id: Optional[str]
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self.frame_ids[id(frame)] = frame_id
        elif event == "return":
            frame_id = self.frame_ids.get(id(frame))
            if frame_id is None:
                frame_id = f"frm_{ulid.new()}"
        else:
            frame_id = None  # pragma: no cover
        return frame_id

    def process(self, frame, event, arg, call_frames):
        if self.events is not None and event not in self.events:
            return None

        timestamp = time.time()
        frame_id = self.frame_id(frame, event)
        data = {
            "frame_id": frame_id,
            "timestamp": timestamp,
            "user_code_call_site": user_code_call_site(call_frames, frame_id),
        }

        if event == "call":
            data["type"] = self.call_type
        elif event == "return":
            data["type"] = self.return_type
        elif event == "unwind":
            data["type"] = self.unwind_type
        elif event == "resume":
            data["type"] = self.resume_type
        elif event == "yield":
            data["type"] = self.yield_type
        elif event == "throw":  # pragma: no branch
            data["type"] = self.throw_type

        if self.subtype is not None:
            data["subtype"] = self.subtype

        if self.process_extra is not None:
            data.update(self.process_extra(frame, event, arg, self.context))
        return data

    def __repr__(self):
        return f"PluginProcessor({self.plugin_data})"



import dis
import functools
import logging
import operator
import os
import platform
import sys
import threading
import time
from collections import defaultdict
from typing import Optional

import httpx
import msgpack
import ulid

from .config import CONFIG_KEYS_TO_OMIT_FROM_SAVED_TRACE
from .db import save_trace_in_sqlite
from .filters.core import LibraryPathFilter
from .filters.kolo import kolo_filter_filename
from .git import COMMIT_SHA
from .plugins import PluginProcessor, load_plugin_data
from .serialize import dump_msgpack, dump_msgpack_lightweight_repr, user_code_call_site
from .threads import get_thread_id
from .upload import upload_to_dashboard
from .utils import extract_http_trace_name
from .version import __version__

logger = logging.getLogger("kolo")


def frozen_filter(filename):
    return "<frozen " in filename


def pypy_filter(filename):
    return "<builtin>/" in filename


def exec_filter(filename):
    return filename == "<string>"


def attrs_filter(filename):
    if filename:
        return filename.startswith("<attrs generated")
    else:  # pragma: no cover
        # Index of the frame in the Python stack:
        # 0: This function
        # 1: KoloMonitor.ignore method
        # 2: KoloMonitor.monitor_ method
        # 3: Attrs code with empty filename
        # 4: Attrs code that may match attr/_make.py
        # 5: Attrs code that may match attr/_make.py
        frame = sys._getframe(4)
        if frame is not None and frame.f_code.co_filename == "":
            frame = sys._getframe(5)
        return frame is not None and frame.f_code.co_filename.endswith(
            os.path.normpath("attr/_make.py")
        )


def pytest_filter(filename):
    return filename == "<pytest match expression>"


def build_frame_data(
    event, name, filename, frame, frame_id, call_site, omit_return_locals=False
):
    if omit_return_locals:
        frame_locals = None
    else:
        frame_locals = {k: v for k, v in frame.f_locals.items() if k != "__builtins__"}
    return {
        "path": frame_path(filename, frame.f_lineno),
        "co_name": name,
        "qualname": get_qualname(frame),
        "event": event,
        "frame_id": frame_id,
        "locals": frame_locals,
        "timestamp": time.time(),
        "type": "frame",
        "user_code_call_site": call_site,
    }


def frame_path(path, lineno):
    try:
        relative_path = os.path.relpath(path)
    except ValueError:
        relative_path = path
    return f"{relative_path}:{lineno}"


def get_qualname(frame):
    qualname = frame.f_code.co_qualname
    module = frame.f_globals.get("__name__", "<unknown>")
    return f"{module}.{qualname}"


class KoloLocals(threading.local):
    def __init__(self):
        self.call_frames = []
        self._frame_ids = {}
        self.line_frame = None
        self.line_frame_data = None
        self.variable = None
        self.opname = None


def get_instruction(code, offset):  # pragma: no cover
    for instruction in dis.get_instructions(code):
        if instruction.offset == offset:
            return instruction
    return None


if sys.version_info >= (3, 12):
    PY_START = sys.monitoring.events.PY_START
    PY_RETURN = sys.monitoring.events.PY_RETURN
    PY_UNWIND = sys.monitoring.events.PY_UNWIND
    PY_RESUME = sys.monitoring.events.PY_RESUME
    PY_YIELD = sys.monitoring.events.PY_YIELD
    PY_THROW = sys.monitoring.events.PY_THROW
    INSTRUCTION = sys.monitoring.events.INSTRUCTION
    NO_EVENTS = sys.monitoring.events.NO_EVENTS

    class KoloMonitor:
        def __init__(
            self,
            db_path,
            *,
            config=None,
            one_trace_per_test=False,
            source,
            name: Optional[str] = None,
        ):
            self.tool_id = sys.monitoring.PROFILER_ID
            self.active = False
            self.timestamp = None
            self.db_path = db_path
            self.config = config if config is not None else {}
            self.source = source
            self.one_trace_per_test = one_trace_per_test
            self.omit_return_locals = self.config.get("omit_return_locals", False)
            self.trace_id = f"trc_{ulid.new()}"
            self.trace_name = name

            filters = self.config.get("filters", {})
            self.include_frames = [
                os.path.normpath(f) for f in filters.get("include_frames", [])
            ]
            self.ignore_frames = [
                os.path.normpath(f) for f in filters.get("ignore_frames", [])
            ]

            self.default_include_frames = {}
            for plugin_data in load_plugin_data(self.config):
                processor = PluginProcessor(plugin_data, self.config)
                for co_name in plugin_data["co_names"]:
                    self.default_include_frames.setdefault(co_name, []).append(
                        processor
                    )

            self.default_ignore_frames = [
                LibraryPathFilter().filter,
                frozen_filter,
                pypy_filter,
                kolo_filter_filename,
                exec_filter,
                attrs_filter,
                pytest_filter,
            ]
            self.thread_locals = KoloLocals()
            self.line_events = self.config.get("line_events", False)

            # Key is the thread id, value is the native python thread object
            self.threads = {}

            # Key is the thread id, value is a list of frames for that thread
            self.frames_by_thread = defaultdict(list)

            self.current_thread_id = get_thread_id(threading.current_thread())
            if self.config.get("lightweight_repr", False):
                self.dump_msgpack = dump_msgpack_lightweight_repr
            else:
                self.dump_msgpack = dump_msgpack

        def ignore(self, filename):
            for ignore in self.default_ignore_frames:
                if ignore(filename):
                    return True
            return False

        def include(self, processor, event, filename, name, arg):
            try:
                if not processor.matches(filename, name):
                    return None

                # Index of the frame in the Python stack:
                # 0: This method
                # 1: KoloMonitor.monitor_ method
                # 2: The frame we want
                frame = sys._getframe(2)
                if processor.call_extra is not None and not processor.call_extra(
                    frame, event, arg, processor.context
                ):
                    return None

                return processor.process(
                    frame, event, arg, self.thread_locals.call_frames
                )
            except Exception as e:
                logger.warning(
                    "Unexpected exception in default_include_frames: %s",
                    processor,
                    exc_info=e,
                )
                return None

        def start_test(self):
            self.trace_id = f"trc_{ulid.new()}"
            self.start_test_indices = {
                thread_id: len(frames)
                for thread_id, frames in self.frames_by_thread.items()
            }

        def end_test(self):
            frames_by_thread = {
                thread_id: frames[self.start_test_indices.get(thread_id, 0) :]
                for thread_id, frames in self.frames_by_thread.items()
            }
            self.save(frames_by_thread=frames_by_thread)

        def monitor_pystart(self, code, instruction_offset):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            filename = code.co_filename
            name = code.co_name

            frames = []
            frame_types = []

            try:
                if name in self.default_include_frames:
                    processors = self.default_include_frames[name]
                    for processor in processors:
                        frame_data = self.include(
                            processor, "call", filename, name, None
                        )
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append(frame_data["type"])

                for path in self.include_frames:
                    if path in filename:
                        frames.append(self.process_pystart(filename, name))
                        frame_types.append("frame")
                        return

                if self.ignore(filename):
                    if frames:
                        return  # Don't disable if a default processor matched
                    return sys.monitoring.DISABLE

                for path in self.ignore_frames:
                    if path in filename:
                        if frames:
                            return  # Don't disable if a default processor matched
                        return sys.monitoring.DISABLE

                frames.append(self.process_pystart(filename, name))
                frame_types.append("frame")
            finally:
                self.push_frames_call(frames, frame_types)

        def monitor_pyreturn(
            self, code, instruction_offset, retval
        ):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            filename = code.co_filename
            name = code.co_name

            frames = []
            frame_types = []

            try:
                if name in self.default_include_frames:
                    processors = self.default_include_frames[name]
                    for processor in processors:
                        frame_data = self.include(
                            processor, "return", filename, name, retval
                        )
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append(frame_data["type"])

                for path in self.include_frames:
                    if path in filename:
                        frames.append(self.process_pyreturn(filename, name, retval))
                        frame_types.append("frame")
                        return

                if self.ignore(filename):
                    if frames:
                        return  # Don't disable if a default processor matched
                    return sys.monitoring.DISABLE

                for path in self.ignore_frames:
                    if path in filename:
                        if frames:
                            return  # Don't disable if a default processor matched
                        return sys.monitoring.DISABLE

                frames.append(self.process_pyreturn(filename, name, retval))
                frame_types.append("frame")
            finally:
                self.push_frames_return(frames, frame_types)

        def monitor_pyunwind(
            self, code, instruction_offset, exception
        ):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            filename = code.co_filename
            name = code.co_name

            frames = []
            frame_types = []

            try:
                if name in self.default_include_frames:
                    processors = self.default_include_frames[name]
                    for processor in processors:
                        frame_data = self.include(
                            processor, "unwind", filename, name, exception
                        )
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append(frame_data["type"])

                for path in self.include_frames:
                    if path in filename:
                        frames.append(self.process_pyunwind(filename, name, exception))
                        frame_types.append("frame")
                        return

                if self.ignore(filename):
                    # We would like to return `sys.monitoring.DISABLE` here, but
                    # `PY_UNWIND` events cannot be disabled, so we do the next best
                    # thing and return asap.
                    return

                for path in self.ignore_frames:
                    if path in filename:
                        # We would like to return `sys.monitoring.DISABLE` here, but
                        # `PY_UNWIND` events cannot be disabled, so we do the next best
                        # thing and return asap.
                        return

                frames.append(self.process_pyunwind(filename, name, exception))
                frame_types.append("frame")
            finally:
                self.push_frames_return(frames, frame_types)

        def monitor_pyresume(self, code, instruction_offset):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            filename = code.co_filename
            name = code.co_name

            frames = []
            frame_types = []

            try:
                if name in self.default_include_frames:
                    processors = self.default_include_frames[name]
                    for processor in processors:
                        frame_data = self.include(
                            processor, "resume", filename, name, None
                        )
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append(frame_data["type"])

                for path in self.include_frames:
                    if path in filename:
                        frames.append(self.process_pyresume(filename, name))
                        frame_types.append("frame")
                        return

                if self.ignore(filename):
                    if frames:
                        return  # Don't disable if a default processor matched
                    return sys.monitoring.DISABLE

                for path in self.ignore_frames:
                    if path in filename:
                        if frames:
                            return  # Don't disable if a default processor matched
                        return sys.monitoring.DISABLE

                frames.append(self.process_pyresume(filename, name))
                frame_types.append("frame")
            finally:
                self.push_frames_call(frames, frame_types)

        def monitor_pyyield(self, code, instruction_offset, retval):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            filename = code.co_filename
            name = code.co_name

            frames = []
            frame_types = []

            try:
                if name in self.default_include_frames:
                    processors = self.default_include_frames[name]
                    for processor in processors:
                        frame_data = self.include(
                            processor, "yield", filename, name, retval
                        )
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append(frame_data["type"])

                for path in self.include_frames:
                    if path in filename:
                        frame_data = self.process_pyyield(filename, name, retval)
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append("frame")
                            return

                if self.ignore(filename):
                    if frames:
                        return  # Don't disable if a default processor matched
                    return sys.monitoring.DISABLE

                for path in self.ignore_frames:
                    if path in filename:
                        if frames:
                            return  # Don't disable if a default processor matched
                        return sys.monitoring.DISABLE

                frame_data = self.process_pyyield(filename, name, retval)
                if frame_data is not None:
                    frames.append(frame_data)
                    frame_types.append("frame")
            finally:
                self.push_frames_return(frames, frame_types)

        def monitor_pythrow(
            self, code, instruction_offset, exception
        ):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            filename = code.co_filename
            name = code.co_name

            frames = []
            frame_types = []

            try:
                if name in self.default_include_frames:
                    processors = self.default_include_frames[name]
                    for processor in processors:
                        frame_data = self.include(
                            processor, "throw", filename, name, exception
                        )
                        if frame_data is not None:
                            frames.append(frame_data)
                            frame_types.append(frame_data["type"])

                for path in self.include_frames:
                    if path in filename:
                        frames.append(self.process_pythrow(filename, name, exception))
                        frame_types.append("frame")
                        return

                if self.ignore(filename):
                    # We would like to return `sys.monitoring.DISABLE` here, but
                    # `PY_UNWIND` events cannot be disabled, so we do the next best
                    # thing and return asap.
                    return

                for path in self.ignore_frames:
                    if path in filename:
                        # We would like to return `sys.monitoring.DISABLE` here, but
                        # `PY_UNWIND` events cannot be disabled, so we do the next best
                        # thing and return asap.
                        return

                frames.append(self.process_pythrow(filename, name, exception))
                frame_types.append("frame")
            finally:
                self.push_frames_call(frames, frame_types)

        def monitor_instruction(self, code, instruction_offset):  # pragma: no cover
            if self.thread_locals.opname is not None:
                self.process_assignment()

            instruction = get_instruction(code, instruction_offset)

            if instruction is None or instruction.opname not in (
                "STORE_FAST",
                "STORE_GLOBAL",
                "STORE_DEREF",
            ):
                return sys.monitoring.DISABLE

            # pytest assert internals
            argval = instruction.argval
            if argval is not None and instruction.argval.startswith("@"):
                return sys.monitoring.DISABLE

            filename = code.co_filename
            name = code.co_name

            for path in self.include_frames:
                if path in filename:
                    self.process_instruction(filename, name, instruction)
                    return

            if self.ignore(filename):
                return sys.monitoring.DISABLE

            for path in self.ignore_frames:
                if path in filename:
                    return sys.monitoring.DISABLE

            self.process_instruction(filename, name, instruction)

        def process_pystart(self, filename, name):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_pystart method
            # 2: The frame we want
            frame = sys._getframe(2)
            frame_id = f"frm_{ulid.new()}"
            self.thread_locals._frame_ids[id(frame)] = frame_id

            user_code_call_site_ = user_code_call_site(
                self.thread_locals.call_frames, frame_id
            )

            self.thread_locals.call_frames.append((frame, frame_id))

            data = build_frame_data(
                "call",
                name,
                filename,
                frame,
                frame_id,
                user_code_call_site_,
            )
            data["arg"] = None
            return data

        def process_pyreturn(self, filename, name, retval):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_pyreturn method
            # 2: The frame we want
            frame = sys._getframe(2)
            frame_id = self.thread_locals._frame_ids[id(frame)]

            user_code_call_site_ = user_code_call_site(
                self.thread_locals.call_frames, frame_id
            )

            self.thread_locals.call_frames.pop()

            data = build_frame_data(
                "return",
                name,
                filename,
                frame,
                frame_id,
                user_code_call_site_,
                self.omit_return_locals,
            )
            data["arg"] = retval
            return data

        def process_pyunwind(self, filename, name, exception):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_pyunwind method
            # 2: The frame we want
            frame = sys._getframe(2)
            frame_id = self.thread_locals._frame_ids[id(frame)]

            user_code_call_site_ = user_code_call_site(
                self.thread_locals.call_frames, frame_id
            )

            self.thread_locals.call_frames.pop()

            data = build_frame_data(
                "unwind",
                name,
                filename,
                frame,
                frame_id,
                user_code_call_site_,
            )
            data["exception"] = exception
            return data

        def process_pyresume(self, filename, name):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_pyresume method
            # 2: The frame we want
            frame = sys._getframe(2)
            frame_id = f"frm_{ulid.new()}"
            self.thread_locals._frame_ids[id(frame)] = frame_id

            user_code_call_site_ = user_code_call_site(
                self.thread_locals.call_frames, frame_id
            )

            self.thread_locals.call_frames.append((frame, frame_id))

            data = build_frame_data(
                "resume",
                name,
                filename,
                frame,
                frame_id,
                user_code_call_site_,
            )
            data["arg"] = None
            return data

        def process_pyyield(self, filename, name, retval):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_pyyield method
            # 2: The frame we want
            frame = sys._getframe(2)
            try:
                frame_id = self.thread_locals._frame_ids[id(frame)]
            except KeyError:
                # We enabled Kolo in a generator or async function, so missed the start/resume frame
                return

            user_code_call_site_ = user_code_call_site(
                self.thread_locals.call_frames, frame_id
            )

            self.thread_locals.call_frames.pop()

            data = build_frame_data(
                "yield",
                name,
                filename,
                frame,
                frame_id,
                user_code_call_site_,
                self.omit_return_locals,
            )
            data["arg"] = retval
            return data

        def process_pythrow(self, filename, name, exception):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_pythrow method
            # 2: The frame we want
            frame = sys._getframe(2)
            frame_id = f"frm_{ulid.new()}"
            self.thread_locals._frame_ids[id(frame)] = frame_id

            user_code_call_site_ = user_code_call_site(
                self.thread_locals.call_frames, frame_id
            )

            self.thread_locals.call_frames.append((frame, frame_id))

            data = build_frame_data(
                "throw",
                name,
                filename,
                frame,
                frame_id,
                user_code_call_site_,
            )
            data["exception"] = exception
            return data

        def process_instruction(self, filename, name, instruction):  # pragma: no cover
            # Index of the frame in the Python stack:
            # 0: This function
            # 1: KoloMonitor.monitor_instruction method
            # 2: The frame we want
            frame = sys._getframe(2)
            frame_id = self.thread_locals._frame_ids.get(id(frame))

            self.thread_locals.variable = instruction.argval
            self.thread_locals.opname = instruction.opname
            self.thread_locals.line_frame = frame
            self.thread_locals.line_frame_data = {
                "path": frame_path(filename, frame.f_lineno),
                "co_name": name,
                "qualname": get_qualname(frame),
                "event": "line",
                "frame_id": frame_id,
                "timestamp": time.time(),
                "type": "frame",
            }

        def process_assignment(self):  # pragma: no cover
            frame = self.thread_locals.line_frame
            variable = self.thread_locals.variable
            opname = self.thread_locals.opname
            frame_data = self.thread_locals.line_frame_data

            if opname == "STORE_FAST":
                frame_data["assign"] = (variable, frame.f_locals[variable])
            elif opname == "STORE_GLOBAL":
                frame_data["assign"] = (variable, frame.f_globals[variable])
            elif opname == "STORE_DEREF":  # nonlocal
                frame_data["assign"] = (variable, frame.f_locals[variable])

            self.push_frame_data(frame_data)

            self.thread_locals.line_frame = None
            self.thread_locals.variable = None
            self.thread_locals.opname = None
            self.thread_locals.line_frame_data = None

        def push_frame_data(self, data):
            data = self.dump_msgpack(data)

            current_thread = threading.current_thread()
            thread_id = get_thread_id(current_thread)
            if thread_id not in self.threads:
                self.threads[thread_id] = current_thread

            self.frames_by_thread[thread_id].append(data)

        def push_frames_call(self, frames, frame_types):
            if not frames:
                pass
            elif frame_types == ["frame"]:
                self.push_frame_data(frames[0])
            else:
                if self.one_trace_per_test:
                    for index, frame_type in enumerate(frame_types):
                        if frame_type == "start_test":
                            before, frames = frames[:index], frames[index:]
                            for frame_data in before:
                                self.push_frame_data(frame_data)

                            self.start_test()

                for frame_data in frames:
                    self.push_frame_data(frame_data)

        def push_frames_return(self, frames, frame_types):
            if not frames:
                pass
            elif frame_types == ["frame"]:
                self.push_frame_data(frames[0])
            else:
                frames.reverse()
                frame_types.reverse()

                if self.one_trace_per_test:
                    for index, frame_type in enumerate(frame_types):
                        if frame_type == "end_test":
                            before, frames = frames[: index + 1], frames[index + 1 :]
                            for frame_data in before:
                                self.push_frame_data(frame_data)

                            self.end_test()

                for frame_data in frames:
                    self.push_frame_data(frame_data)

        def build_trace(self, frames_by_thread=None):
            # frames_by_thread is only passed when called by end_test.

            config = {
                k: v
                for k, v in self.config.items()
                if k not in CONFIG_KEYS_TO_OMIT_FROM_SAVED_TRACE
            }
            # Config to ALWAYS include, even if the user user didn't change the default
            config["use_monitoring"] = True
            config["use_rust"] = False

            data = {
                "command_line_args": sys.argv,
                "current_commit_sha": COMMIT_SHA,
                "current_thread_id": self.current_thread_id,
                "meta": {
                    "version": __version__,
                    "source": self.source,
                    "environment": {
                        "py_version": platform.python_version(),
                        "py_version_full": sys.version,
                        "platform": platform.platform(),
                        "system": platform.system(),
                        "machine": platform.machine(),
                        "processor": platform.processor(),
                    },
                    "config": config,
                },
                "timestamp": self.timestamp,
                "trace_id": self.trace_id,
                "trace_name": self.trace_name,
            }

            if frames_by_thread is None:
                frames_by_thread = self.frames_by_thread

            # frames_of_interest and frames are always empty
            # for sys.monitoring traces now.
            data["frames_of_interest"] = []
            data["frames"] = {}

            data["threads"] = {}

            for thread_id, thread in self.threads.items():
                data["threads"][thread_id] = {
                    "name": thread.name,
                    "ident": getattr(thread, "ident", None),
                    "native_id": getattr(thread, "native_id", None),
                    "daemon": thread.daemon,
                    "is_alive": thread.is_alive(),
                }

                if thread_id in frames_by_thread:
                    data["threads"][thread_id]["frames"] = [
                        msgpack.unpackb(f, strict_map_key=False)
                        for f in frames_by_thread[thread_id]
                    ]

            # TODO: And then let's also capture threads with threading.enumerate()
            # just for some fun extra data...
            return self.dump_msgpack(data)

        def save(self, frames_by_thread=None):
            """
            frames_by_thread is only passed when called from end_test,
            because end_test cuts off some frames and saves directly.
            """

            if self.trace_name is None:
                self._set_http_trace_name(frames_by_thread)

            serialized_data = self.build_trace(frames_by_thread=frames_by_thread)
            timeout = self.config.get("sqlite_busy_timeout", 60)
            save_trace_in_sqlite(
                self.db_path, self.trace_id, msgpack=serialized_data, timeout=timeout
            )

        def _set_http_trace_name(self, frames_by_thread=None):
            """
            Extract HTTP request/response information from frames to set the trace name.
            """
            if frames_by_thread is None:
                frames_by_thread = self.frames_by_thread

            trace_name = extract_http_trace_name(
                frames_by_thread, self.current_thread_id
            )
            if trace_name:
                self.trace_name = trace_name

    def activate_monitoring(monitor):
        monitor.active = True
        tool_id = monitor.tool_id
        sys.monitoring.use_tool_id(tool_id, "kolo")
        all_events = PY_START | PY_RETURN | PY_UNWIND | PY_RESUME | PY_YIELD | PY_THROW

        sys.monitoring.register_callback(tool_id, PY_START, monitor.monitor_pystart)
        sys.monitoring.register_callback(tool_id, PY_RETURN, monitor.monitor_pyreturn)
        sys.monitoring.register_callback(tool_id, PY_UNWIND, monitor.monitor_pyunwind)
        sys.monitoring.register_callback(tool_id, PY_RESUME, monitor.monitor_pyresume)
        sys.monitoring.register_callback(tool_id, PY_YIELD, monitor.monitor_pyyield)
        sys.monitoring.register_callback(tool_id, PY_THROW, monitor.monitor_pythrow)

        if monitor.line_events:
            all_events |= INSTRUCTION
            sys.monitoring.register_callback(
                tool_id, INSTRUCTION, monitor.monitor_instruction
            )

        sys.monitoring.set_events(tool_id, all_events)

        monitor.timestamp = time.time()

    def disable_monitoring(monitor):
        if monitor.active:
            monitor.active = False
            sys.monitoring.set_events(monitor.tool_id, NO_EVENTS)
            sys.monitoring.register_callback(monitor.tool_id, PY_START, None)
            sys.monitoring.register_callback(monitor.tool_id, PY_RETURN, None)
            sys.monitoring.register_callback(monitor.tool_id, PY_UNWIND, None)
            sys.monitoring.register_callback(monitor.tool_id, PY_RESUME, None)
            sys.monitoring.register_callback(monitor.tool_id, PY_YIELD, None)
            sys.monitoring.register_callback(monitor.tool_id, PY_THROW, None)
            if monitor.line_events:
                sys.monitoring.register_callback(monitor.tool_id, INSTRUCTION, None)
            sys.monitoring.free_tool_id(monitor.tool_id)
            sys.monitoring.restart_events()



from __future__ import annotations

import platform
from datetime import datetime, timezone

from ..config import load_config
from ..db import load_trace_from_db, setup_db
from ..serialize import load_msgpack
from ..utils import extract_main_frames_from_data, maybe_format
from ..version import __version__
from .format import format_intermediate
from .plan import Plan, build_steps, load_hooks, run_plan_hooks, run_step_hooks
from .processors import load_processors, run_processors


def load_traces(db_path, trace_ids):
    traces = {}
    for trace_id in trace_ids:
        raw_msgpack, _ = load_trace_from_db(db_path, trace_id)
        trace = load_msgpack(raw_msgpack)

        # TODO: We could do something fancier with other threads here..
        frames = extract_main_frames_from_data(trace)

        traces[trace_id] = {
            "frames": frames,
            "trace": trace,
        }
    return traces


def build_test_context(
    *trace_ids: str,
    test_class: str,
    test_name: str,
    config,
    include_generation_timestamp=True,
    use_saved_schemas=False,
):
    processors = load_processors(config)

    db_path = setup_db()
    traces = load_traces(db_path, trace_ids)

    context = {
        "_config": config,
        "_db_path": db_path,
        "_traces": traces,
        "_use_saved_schemas": use_saved_schemas,
        "kolo_version": __version__,
        "now": datetime.now(timezone.utc) if include_generation_timestamp else None,
        "test_class": test_class,
        "test_name": test_name,
    }
    run_processors(processors, context)
    return context


def create_test_plan(config, context, pytest=True) -> Plan:
    plan = Plan(build_steps(context, pytest=pytest), pytest, context)
    plan_hooks, step_hooks = load_hooks(config)
    plan = run_plan_hooks(plan, plan_hooks)
    return run_step_hooks(plan, step_hooks)


def generate_from_trace_ids(
    *trace_ids: str,
    test_class: str,
    test_name: str,
    config=None,
    include_generation_timestamp=True,
    use_saved_schemas=False,
    pytest=True,
) -> str:
    if config is None:
        config = load_config()
    context = build_test_context(
        *trace_ids,
        test_class=test_class,
        test_name=test_name,
        config=config,
        include_generation_timestamp=include_generation_timestamp,
        use_saved_schemas=use_saved_schemas,
    )
    plan = create_test_plan(config, context, pytest)
    rendered = plan.render()
    return maybe_format(rendered)


def generate_test_intermediate_format(
    *trace_ids: str,
    test_class: str,
    test_name: str,
    config=None,
    include_generation_timestamp=True,
    use_saved_schemas: bool = False,
):
    if config is None:  # pragma: no branch
        config = load_config()
    processors = load_processors(config)

    db_path = setup_db()
    traces = load_traces(db_path, trace_ids)

    context = {
        "_config": config,
        "_db_path": db_path,
        "_traces": traces,
        "_use_saved_schemas": use_saved_schemas,
        "kolo_version": __version__,
        "now": datetime.now(timezone.utc) if include_generation_timestamp else None,
        "test_class": test_class,
        "test_name": test_name,
    }

    for processor in processors:  # pragma: no branch
        output = processor(context)
        if output:
            context.update(output)
        if "sections" in context and "sql_fixtures" in context["sections"][0]:
            yield from format_intermediate(context)
            break
